from .data_helpers import *
from .data_spec import DataSpec
from .variant_frequencies import VariantFrequencies
from .hier_frequencies import HierFrequencies
from .case_counts import CaseCounts
from .hier_cases import HierCases
from .case_frequencies import CaseFrequencyData, HierarchicalCFData



from functools import partial
from typing import Optional
import numpy as np
from jax import vmap
import jax.numpy as jnp
from jax.nn import softmax

import numpyro
import numpyro.distributions as dist
from numpyro.infer.reparam import TransformReparam

from .model_spec import ModelSpec
from .multinomial_logistic_regression import (
    MultinomialLogisticRegression,
    simulate_MLR,
)


def simulate_hier_mlr(growth_advantages, freq0, tau, Ns):
    # Assume all are in list of size groups
    assert len(growth_advantages) == len(freq0)
    assert len(growth_advantages) == len(Ns)
    groups = len(growth_advantages)
    seq_counts = []
    for group in range(groups):
        _, sc_group = simulate_MLR(
            growth_advantages[group], freq0[group], tau, Ns[group]
        )
        seq_counts.append(sc_group)
    return seq_counts


def hier_MLR_numpyro(
    seq_counts, N, X, tau=None, pool_scale=None, pred=False, var_names=None
):
    _, N_variants, N_groups = seq_counts.shape
    _, N_features, _ = X.shape

    pool_scale = 0.1 if pool_scale is None else pool_scale

    # Sampling intercept and fitness parameters
    reparam_config = {
        "beta_loc": TransformReparam(),
        "beta_scale": TransformReparam(),
        "alpha": TransformReparam(),
        "raw_beta": TransformReparam(),
    }
    with numpyro.handlers.reparam(config=reparam_config):
        beta_scale = numpyro.sample(
            "beta_scale",
            dist.TransformedDistribution(
                dist.HalfNormal(1.0),
                dist.transforms.AffineTransform(0.0, pool_scale),
            ),
        )
        with numpyro.plate("variants", N_variants - 1, dim=-2):
        # Define loc and scale for fitness beta
            beta_loc = numpyro.sample(
                "beta_loc",
                dist.TransformedDistribution(
                    dist.Normal(0.0, 1.0),
                    dist.transforms.AffineTransform(0.0, 0.2),
                ),
            )

            # Use location and scale parameter to draw within group
            with numpyro.plate("group", N_groups, dim=-1):
                # Leave intercept alpha unpooled
                raw_alpha = numpyro.sample(
                    "alpha",
                    dist.TransformedDistribution(
                        dist.Normal(0.0, 1.0),
                        dist.transforms.AffineTransform(0.0, 6.0),
                    ),
                )
                raw_beta = numpyro.sample(
                    "raw_beta",
                    dist.TransformedDistribution(
                        dist.Normal(0.0, 1.0),
                        dist.transforms.AffineTransform(beta_loc, beta_scale),
                    ),
                )

    # All parameters are relative to last column / variant
    beta = numpyro.deterministic(
        "beta",
        jnp.concatenate(
            (
                jnp.stack((raw_alpha, raw_beta)),
                jnp.zeros((N_features, 1, N_groups)),
            ),
            axis=1,
        ),
    )

    # (T, F, G) times (F, V, G) -> (T, V, G)
    dot_by_group = vmap(jnp.dot, in_axes=(-1, -1), out_axes=-1)
    logits = dot_by_group(X, beta)  # Logit frequencies by variant

    # Evaluate likelihood
    obs = None if pred else np.swapaxes(np.nan_to_num(seq_counts), 1, 2)

    _seq_counts = numpyro.sample(
        "_seq_counts",
        dist.MultinomialLogits(
            logits=jnp.swapaxes(logits, 1, 2), total_count=np.nan_to_num(N)
        ),
        obs=obs,
    )

    # Re-ordering so groups are last
    seq_counts = numpyro.deterministic(
        "seq_counts", jnp.swapaxes(_seq_counts, 2, 1)
    )

    # Compute frequency
    numpyro.deterministic("freq", softmax(logits, axis=1))

    # Compute growth advantage from model
    if tau is not None:
        numpyro.deterministic(
            "ga", jnp.exp(beta[-1, :-1, :] * tau)
        )  # Last row corresponds to linear predictor / growth advantage
        numpyro.deterministic("ga_loc", jnp.exp(beta_loc[:, 0] * tau))


class HierMLR(ModelSpec):
    def __init__(self, tau: float, pool_scale: Optional[float] = None) -> None:
        """Construct ModelSpec for Hierarchial multinomial logistic regression.

        Parameters
        ----------
        tau:
            Assumed generation time for conversion to relative R.

        pool_scale:
            Prior standard deviation for pooling of growth advantages.

        Returns
        -------
        HierMLR
        """
        self.tau = tau  # Fixed generation time
        self.pool_scale = pool_scale  # Prior std for coefficients
        self.model_fn = partial(hier_MLR_numpyro, pool_scale=self.pool_scale)

    @staticmethod
    def make_ols_feature(start, stop, n_groups):
        """
        Construct simple OLS features (1, x) for HierMLR as nd.array.

        Parameters
        ----------
        start:
            Start value for OLS feature.
        stop:
            Stop value for OLS feature.

        n_groups:
            number of groups in the hierarchical model.
        """
        X_flat = MultinomialLogisticRegression.make_ols_feature(start, stop)
        return np.stack([X_flat] * n_groups, axis=-1)

    def augment_data(self, data: dict) -> None:
        T, G = data["N"].shape
        data["tau"] = self.tau
        data["X"] = self.make_ols_feature(0, T, G)

    @staticmethod
    def forecast_frequencies(samples, forecast_L):
        """
        Use posterior beta to forecast posterior frequencies.
        """

        # Making feature matrix for forecasting
        last_T = samples["freq"].shape[1]
        n_groups = samples["freq"].shape[-1]

        X = HierMLR.make_ols_feature(
            start=last_T, stop=last_T + forecast_L, n_groups=n_groups
        )

        # (T, F, G) times (F, V, G) -> (T, V, G)
        dot_by_group = vmap(jnp.dot, in_axes=(-1, -1), out_axes=-1)
        dbg_by_sample = vmap(dot_by_group, in_axes=(None, 0), out_axes=0)

        # Creating frequencies from posterior beta
        beta = jnp.array(samples["beta"])
        logits = dbg_by_sample(X, beta)
        samples["freq_forecast"] = softmax(logits, axis=-2)  # (S, T, V, G)
        return samples



from .data_helpers import prep_dates, prep_sequence_counts
from .data_spec import DataSpec
import pandas as pd
from typing import List, Optional

VARIANT_NAMES = ["Variant", "other"]
START_DATE = pd.to_datetime("2022-01-01")


def variant_counts_to_dataframe(
    var_counts,
    var_names: List[str] = VARIANT_NAMES,
    start_date=START_DATE,
):
    """Convert matrix of variant counts to pandas dataframe
    for input to ef.VariantFrequencies.

    Parameters
    ----------
    var_counts:
        nd.array of counts var_counts[t,v] of variant v on day t.
    variant_names:
        List of variant names to assign each column.
    start_date:
        Pandas datetime to use as first date.

    Returns
    -------
    seq_counts
    """
    T, V = var_counts.shape

    assert V <= len(var_names), "More cols present in var_counts than names"

    sequences = []
    variant = []
    dates = []

    for time in range(T):
        # Add wildtype and variant counts
        sequences.extend(list(var_counts[time, :]))

        # Add variant labels
        variant.extend(var_names[:V])

        # Add dates
        dates.extend([start_date + pd.to_timedelta(time, unit="d")] * V)

    df = pd.DataFrame(
        {
            "variant": variant,
            "sequences": sequences,
            "date": dates,
        }
    )

    return df[df.sequences > 0].reset_index(drop=True)


class VariantFrequencies(DataSpec):
    def __init__(
        self,
        raw_seq: pd.DataFrame,
        date_to_index: Optional[dict] = None,
        var_names: Optional[List] = None,
        pivot: Optional[str] = None,
    ):
        """Construct a data specification for handling variant frequencies.

        Parameters
        ----------
        raw_seq:
            a dataframe containing sequence counts with columns 'sequences',
            'variant', and date'.

        date_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        var_names:
            optional list containing names of variants to be present.

        pivot:
            optional name of variant to place last.
            Defaults to "other" if present otherwise.
            This will usually used as a reference or pivot strain.

        Returns
        -------
        VariantFrequencies
        """

        # Get mapping from date to index
        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_seq["date"])
        self.date_to_index = date_to_index

        # Turn dataframe to counts of each variant sequenced each day
        self.var_names, self.seq_counts = prep_sequence_counts(
            raw_seq, self.date_to_index, var_names, pivot=pivot
        )
        self.pivot = self.var_names[-1]

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()

        # Export `seq_counts`, sequences each day, and variant names
        data["seq_counts"] = self.seq_counts
        data["N"] = self.seq_counts.sum(axis=1)
        data["var_names"] = self.var_names
        return data



from abc import ABC, abstractmethod
from typing import Callable


class ModelSpec(ABC):
    """Abstract model class.
    Used by evofr to handle model specifications for inference.
    Classes which inherit from ModelSpec must have an attribute 'model_fn'
    which defines the function to be used for inference in numpyro.
    """
    model_fn: Callable

    @abstractmethod
    def augment_data(self, data: dict) -> None:
        """
        Augments existing data for inference with model specific information.

        Parameters
        ----------
        data:
            dictionary which contains arguments to 'model_fn'

        Returns
        -------
        None
        """
        pass



from typing import Optional
from functools import partial
import jax.numpy as jnp
import numpy as np

from evofr.models.model_spec import ModelSpec
from evofr.models.renewal_model.model_helpers import to_survivor_function
from .basis_functions import BasisFunction, Spline

from .model_functions import forward_simulate_I_and_prev, reporting_to_vec
from .model_options import NegBinomCases

import numpyro
import numpyro.distributions as dist


def _single_renewal_model(
    g_rev,
    delays,
    inf_period,
    seed_L,
    forecast_L,
    cases,
    X,
    day_of_week_effect=True,
    CaseLik=None,
    pred=False,
):
    if CaseLik is None:
        CaseLik = NegBinomCases()

    T, k = X.shape
    obs_range = jnp.arange(seed_L, seed_L + T, 1)

    # Effective Reproduction number likelihood
    gam = numpyro.sample("gam", dist.HalfNormal(0.5))
    beta_0 = numpyro.sample("beta_0", dist.Normal(0.0, 0.5))
    with numpyro.plate("N_steps_base", k - 1):
        beta_rw_step = numpyro.sample("beta_rw_step", dist.Laplace()) * gam
        beta_rw = numpyro.deterministic("beta_rw", jnp.cumsum(beta_rw_step))

    # Combine increments and starting position
    beta_rw = jnp.append(jnp.zeros(1), beta_rw)
    beta = beta_0 + beta_rw
    _R = numpyro.deterministic("R", jnp.exp(X @ beta))

    # Add forecasted values of R
    R = _R
    if forecast_L > 0:
        R_forecast = numpyro.deterministic(
            "R_forecast", jnp.vstack((_R[-1],) * forecast_L)
        )
        R = jnp.hstack((_R, R_forecast))

    # Getting initial conditions
    logI0 = numpyro.sample("logI0", dist.Normal()) * 3.0 + 1.0
    I0 = jnp.exp(logI0)
    intros = jnp.zeros((T + seed_L + forecast_L,))
    intros = intros.at[np.arange(seed_L)].set(I0 * jnp.ones(seed_L))

    # Generate day-of-week reporting fraction

    if day_of_week_effect:
        with numpyro.plate("rho_parms", 6):
            rho_logits = numpyro.sample("rho_logits", dist.Normal()) * 5.0
        _rho = jnp.exp(jnp.append(rho_logits, 0.0))
        rho = numpyro.deterministic("rho", _rho / _rho.sum())
    else:
        rho = jnp.ones(7)
    rho_vec = reporting_to_vec(rho, T)

    I_prev, prev = forward_simulate_I_and_prev(
        intros, R, g_rev, delays, inf_period, seed_L
    )
    I_prev = jnp.clip(I_prev, a_min=1e-12, a_max=1e25)

    # Smooth trajectory for plotting
    numpyro.deterministic(
        "I_smooth", jnp.mean(rho_vec) * jnp.take(I_prev, obs_range, axis=0)
    )
    numpyro.deterministic(
        "prev", jnp.mean(rho_vec) * jnp.take(prev, obs_range, axis=0)
    )

    # Compute growth rate assuming I_{t+1} = I_{t} \exp(r_{t})
    numpyro.deterministic(
        "r",
        jnp.diff(
            jnp.log(jnp.take(I_prev, obs_range, axis=0)),
            prepend=jnp.nan,
            axis=0,
        ),
    )

    # Compute expected cases
    numpyro.deterministic(
        "total_smooth_prev",
        jnp.mean(rho_vec) * jnp.take(I_prev, obs_range),
    )
    EC = numpyro.deterministic("EC", jnp.take(I_prev, obs_range) * rho_vec)

    # Evaluate case likelihood
    CaseLik.model(cases, EC, pred=pred)

    if forecast_L > 0:
        I_forecast = numpyro.deterministic(
            "I_smooth_forecast",
            jnp.mean(rho_vec) * I_prev[(seed_L + T) :],
        )
        numpyro.deterministic(
            "r_forecast",
            jnp.diff(jnp.log(I_forecast), prepend=jnp.nan, axis=0),
        )
    return None


class SingleRenewalModel(ModelSpec):
    def __init__(
        self,
        g,
        delays,
        seed_L: int,
        forecast_L: int,
        inf_period=None,
        k: Optional[int] = None,
        CLik=None,
        basis_fn: Optional[BasisFunction] = None,
        day_of_week_effect: bool = True,
    ):
        self.g_rev = jnp.flip(g, axis=-1)
        self.delays = delays
        self.inf_period = (
            to_survivor_function(inf_period)
            if inf_period is not None
            else jnp.ones(self.g_rev.shape[0])
        )
        self.seed_L = seed_L
        self.forecast_L = forecast_L

        # Making basis expansion for Rt
        self.k = k if k else 10
        self.basis_fn = (
            basis_fn if basis_fn else Spline(s=None, order=4, k=self.k)
        )

        # Defining model likelihoods
        self.day_of_week_effect = day_of_week_effect
        self.CLik = CLik
        self.make_model()

    def make_model(self):
        self.model_fn = partial(
            _single_renewal_model,
            g_rev=self.g_rev,
            delays=self.delays,
            inf_period=self.inf_period,
            seed_L=self.seed_L,
            forecast_L=self.forecast_L,
            day_of_week_effect=self.day_of_week_effect,
            CaseLik=self.CLik,
        )

    def augment_data(self, data):
        # Add feature matrix for parameterization of R
        data["X"] = self.basis_fn.make_features(data, T=data["cases"].shape[0])



from .data import DataSpec
from .data import CaseCounts, VariantFrequencies, CaseFrequencyData
from .data import HierCases, HierFrequencies, HierarchicalCFData
from .data.data_helpers import *

from .models import ModelSpec
from .models import MultinomialLogisticRegression, HierMLR
from .models import PianthamModel
from .models import InnovationMLR, InnovationSequenceCounts
from .models.renewal_model import *

from .infer import SVIHandler, MCMCHandler #, BlackJaxHandler
from .infer import (
    InferSVI,
    InferMCMC,
    InferFullRank,
    InferMAP,
    InferNUTS,
    #InferBlackJax,
    init_to_MAP,
)

from .posterior import PosteriorHandler, MultiPosterior
from .posterior import get_median, get_quantile, get_quantiles
from .posterior import (
    get_site_by_variant,
    get_freq,
    get_sites_quantiles_json,
    get_sites_variants_json,
    EvofrEncoder,
    save_json,
)



from abc import ABC, abstractmethod
from typing import Callable, List, Optional
import numpy as np
import pandas as pd

from evofr.data.data_spec import DataSpec
from evofr.data.data_helpers import prep_dates, format_var_names
from evofr.models.renewal_model.model_options import MultinomialSeq
from .multinomial_logistic_regression import MultinomialLogisticRegression
import jax.numpy as jnp
from jax.nn import softmax

import numpyro
import numpyro.distributions as dist

from .model_spec import ModelSpec

from functools import partial
from .renewal_model import Spline


class HazardModel(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def model(self, N_variants, D):
        pass


class LinearHazard(HazardModel):
    def __init__(self):
        pass

    def model(self, N_variants, D):
        # Assume hazard is a linear function
        # This is pooled so that new variants are closer to mean
        a_mean = numpyro.sample("a_mean", dist.Normal(0.0, 1.0)) * 5.0
        b_mean = numpyro.sample("b_mean", dist.Normal(0.0, 1.0)) * 5.0
        with numpyro.plate("variant", N_variants):
            a = numpyro.sample("a", dist.Normal(0.0, 1.0)) * 5.0 + a_mean
            b = numpyro.sample("b", dist.Normal(0.0, 1.0)) * 5.0 + b_mean
        logit_h = a[:, None] * jnp.arange(D - 1) + b[:, None]
        _h = jnp.exp(logit_h) / (1 + jnp.exp(logit_h))
        return _h


class BetaHazard(HazardModel):
    def __init__(self):
        pass

    def model(self, N_variants, D):
        _h_conc = numpyro.sample("_h_conc", dist.InverseGamma(2.0, 2.0))
        with numpyro.plate("delay", D - 1):
            _h_mean = numpyro.sample("_h_mean", dist.Beta(0.1, 0.1))
            with numpyro.plate("variant", N_variants):
                _h = numpyro.sample(
                    "_h",
                    dist.BetaProportion(mean=_h_mean, concentration=_h_conc),
                )
        return _h


class LogitRWHazard(HazardModel):
    def __init__(self):
        pass

    def model(self, N_variants, D):
        h_scale = numpyro.sample("h_scale", dist.HalfNormal(1.0))
        with numpyro.plate("variant", N_variants):
            logit_h = numpyro.sample(
                "logit_h",
                dist.GaussianRandomWalk(scale=h_scale, num_steps=D - 1),
            )
        _h = jnp.exp(logit_h) / (1 + jnp.exp(logit_h))
        return _h


class LogitSplineHazard(HazardModel):
    def __init__(
        self,
        k: Optional[int] = None,
        order: Optional[int] = None,
        pool_scale: Optional[float] = None,
    ):
        self.k = 4 if k is None else k
        self.order = 4 if order is None else order
        self.pool_scale = 0.2 if pool_scale is None else pool_scale

    def model(self, N_variants, D):
        U = Spline(k=self.k, order=self.order).make_features(T=D - 1)
        _, k = U.shape

        # gammma_scale = numpyro.sample("gamma_scale", dist.HalfNormal(0.1))
        with numpyro.plate("basis_function", k):
            gamma_loc = (
                numpyro.sample("gamma_loc", dist.Normal(0.0, 1.0)) * 5.0
            )
            with numpyro.plate("variant", N_variants):
                gamma = numpyro.sample(
                    "gamma", dist.Normal(gamma_loc, self.pool_scale)
                )

        logit_h = numpyro.deterministic("logit_h", jnp.dot(U, gamma.T))
        _h = jnp.exp(logit_h) / (1 + jnp.exp(logit_h))
        return _h.T


def discrete_hazard_to_pmf_cdf(h):
    V, _ = h.shape
    # survival = jnp.exp(jnp.cumsum(jnp.log(1 - h), axis=1))
    survival = jnp.cumprod(1 - h, axis=1)
    survival = jnp.append(survival, jnp.zeros((V, 1)), 1)
    pmf = jnp.diff(1 - survival, axis=1, prepend=0.0)
    return pmf, 1 - survival


def estimate_delay(seq_count_delays, hazard_model):
    _, N_variants, D = seq_count_delays.shape

    # Assume delay distribution is constant over time
    empirical_delays = np.nan_to_num(seq_count_delays).sum(axis=0)
    total_of_variant = np.nan_to_num(empirical_delays).sum(axis=-1)

    # Generate discrete based on specified hazard model
    h = numpyro.deterministic("h", hazard_model(N_variants, D))

    # Get delay distribution
    p_delay, cdf_delay = discrete_hazard_to_pmf_cdf(h)
    p_delay = numpyro.deterministic(
        "p_delay", jnp.clip(p_delay, 1e-12, 1 - 1e-12).T
    )
    cdf_delay = numpyro.deterministic(
        "cdf_delay", jnp.clip(cdf_delay, 1e-12, 1).T
    )

    # Evaluate likelihood of delays
    numpyro.sample(
        "delays",
        dist.Multinomial(total_count=total_of_variant, probs=p_delay.T),
        obs=empirical_delays,
    )

    return p_delay, cdf_delay


def MLR_nowcast_model(
    seq_counts,
    seq_counts_delay,
    N,
    X,
    hazard_model,
    SeqLik,
    tau=None,
    pred=False,
):
    T, N_variants, D = seq_counts_delay.shape
    _, N_features = X.shape

    # Get time since present used to estimate fraction
    # of variant samples observed thus far.
    time_before_now = np.minimum(T - np.arange(T), D) - 1

    # Sampling parameters for growth advantages
    raw_beta = (
        numpyro.sample(
            "raw_beta",
            dist.Normal(0.0, 1.0),
            sample_shape=(N_features, N_variants - 1),
        )
        * jnp.array([10.0, 1.0])[:, None]
    )

    beta = numpyro.deterministic(
        "beta",
        jnp.column_stack(
            (raw_beta, jnp.zeros(N_features))
        ),  # All parameters are relative to last column / variant
    )

    logits = jnp.dot(X, beta)  # Logit frequencies by variant

    # Compute frequency
    _freq = jnp.exp(logits)
    freq = numpyro.deterministic(
        "freq", jnp.clip(_freq / _freq.sum(axis=-1)[:, None], 1e-12, 1 - 1e-12)
    )

    # Estimate delay distribution by variant
    _, cdf_delay = estimate_delay(seq_counts_delay, hazard_model.model)

    # Get adjusted frequency
    frac_delay = numpyro.deterministic(
        "frac_delay", jnp.take(cdf_delay, time_before_now, axis=0)
    )  # (T, V)

    # Correct recent frequencies by probability that
    # collected sequences have been submitted by observation date.
    _nowcast_freq = frac_delay * freq
    nowcast_freq = _nowcast_freq / _nowcast_freq.sum(axis=-1)[:, None]
    nowcast_freq = jnp.clip(nowcast_freq, 1e-16, 1 - 1e-16)

    # Evaluate likelihood of sequence counts given delays
    SeqLik.model(seq_counts, N, nowcast_freq, pred)  # Evaluate likelihood

    # Compute growth advantage from model
    if tau is not None:
        numpyro.deterministic(
            "ga", jnp.exp(beta[-1, :-1] * tau)
        )  # Last row corresponds to linear predictor / growth advantage


class MLRNowcast(ModelSpec):
    def __init__(
        self,
        tau: float,
        hazard_model: Optional[HazardModel] = None,
        SeqLik=None,
    ) -> None:
        """Construct ModelSpec for MLR accounting for submission delay
        differences betwen variants.

        Parameters
        ----------
        tau:
            Assumed generation time for conversion to relative R.

        hazard_model:
            Numpyro model describing the hazard function for submission delay.

        SeqLik:
            Optional sequence likelihood option: MultinomialSeq or
            DirMultinomialSeq. Defaults to MultinomialSeq.

        Returns
        -------
        MLRNowcast
        """
        self.tau = tau  # Fixed generation time
        self.hazard_model = (
            LinearHazard if hazard_model is None else hazard_model
        )
        self.SeqLik = MultinomialSeq() if SeqLik is None else SeqLik
        self.model_fn = partial(
            MLR_nowcast_model,
            hazard_model=self.hazard_model,
            SeqLik=self.SeqLik,
            tau=self.tau,
        )

    def augment_data(self, data: dict) -> None:
        T = len(data["N"])
        data["X"] = MultinomialLogisticRegression.make_ols_feature(
            0, T
        )  # Use intercept and time as predictors


def prep_sequence_counts_delay(
    raw_seqs: pd.DataFrame,
    date_to_index: Optional[dict] = None,
    var_names: Optional[List] = None,
    max_delay: Optional[int] = None,
    pivot: Optional[str] = None
):
    """Process 'raw_seq' data to nd.array including unobserved dates.

    Parameters
    ----------
    raw_seq:
        a dataframe containing sequence counts with
        columns 'sequences' and 'date'.


    date_to_index:
        optional dictionary for mapping calender dates to nd.array indices.

    var_names:
        optional list of variant to count observations.

    pivot:
        optional name of variant to place last.
        Defaults to "other" if present otherwise.
        This will usually used as a reference or pivot strain.

    Returns
    -------
    var_names:
        list of variants counted

    C:
        nd.array containing number of sequences of each variant on each date.
    """

    if var_names is None:
        raw_var_names = list(pd.unique(raw_seqs.variant))
        raw_var_names.sort()
        var_names = format_var_names(raw_var_names, pivot=pivot)

    raw_seqs["date"] = pd.to_datetime(raw_seqs["date"])
    if date_to_index is None:
        _, date_to_index = prep_dates(raw_seqs["date"])

    # Build matrix
    T = len(date_to_index)
    V = len(var_names)
    D = (
        min(raw_seqs["delay"].max(), max_delay)
        if max_delay is not None
        else raw_seqs["delay"].max()
    )
    C = np.full((T, V, D), np.nan)

    # Loop over rows in dataframe to fill matrix
    for i, s in enumerate(var_names):
        for _, row in raw_seqs[raw_seqs.variant == s].iterrows():
            delay = min(row.delay, D)
            C[date_to_index[row.date], i, delay - 1] = row.sequences

    # Loop over matrix rows to correct for zero counts
    for t in range(T):
        if not np.isnan(C[t, :, :]).all():
            # If not all days samples are nan, replace nan with zeros
            np.nan_to_num(C[t, :, :], copy=False)
    return var_names, C


class DelaySequenceCounts(DataSpec):
    def __init__(
        self,
        raw_seq: pd.DataFrame,
        date_to_index: Optional[dict] = None,
        var_names: Optional[List] = None,
        max_delay: Optional[int] = None,
        pivot: Optional[str] = None
    ):
        """Construct a data specification for handling variant frequencies.

        Parameters
        ----------
        raw_seq:
            a dataframe containing sequence counts with columns 'sequences',
            'variant', and date'.

        date_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        var_names:
            optional list containing names of variants to be present

        max_delay:
            optional integer showing the maximum allowed submission delay.

        pivot:
            optional name of variant to place last.
            Defaults to "other" if present otherwise.
            This will usually used as a reference or pivot strain.

        Returns
        -------
        DelaySequenceCounts
        """

        # Get mapping from date to index
        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_seq["date"])
        self.date_to_index = date_to_index

        # Turn dataframe to counts of each variant sequenced each day
        self.var_names, self.seq_counts_delay = prep_sequence_counts_delay(
            raw_seq, self.date_to_index, var_names, max_delay, pivot
        )
        self.pivot = self.var_names[-1]
        self.seq_counts = self.seq_counts_delay.sum(axis=-1)

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()

        # Export `seq_counts`, sequences each day, and variant names
        data["seq_counts_delay"] = self.seq_counts_delay
        data["seq_counts"] = self.seq_counts
        data["N"] = self.seq_counts.sum(axis=-1)
        return data



from functools import partial
from jax import lax, jit, vmap
import jax.numpy as jnp


@partial(jit, static_argnums=3)
def get_infections(I0, R, g_rev, seed_L):
    max_age = len(g_rev)

    I0_vec = jnp.repeat(I0, seed_L)

    @jit
    def _scan_infections(infections, R):
        curr_I = R * jnp.dot(infections, g_rev)  # New I
        return jnp.append(infections[-(max_age - 1) :], curr_I), curr_I

    _, infections = lax.scan(
        _scan_infections, init=jnp.pad(I0_vec, (max_age - seed_L, 0)), xs=R
    )
    return jnp.append(I0_vec, infections)


@partial(jit, static_argnums=3)
def get_infections_intros(m, R, g_rev, seed_L):
    max_age = len(g_rev)

    @jit
    def _scan_infections(infections, xs):
        R, im = xs
        curr_I = R * jnp.dot(infections, g_rev) + im  # New I
        return jnp.append(infections[-(max_age - 1) :], curr_I), curr_I

    _, infections = lax.scan(
        _scan_infections,
        init=jnp.zeros(max_age),
        xs=(jnp.pad(R, (seed_L, 0), constant_values=1.0), m),
    )
    return infections


@jit
def apply_delay(infections, delay):
    I_padded = jnp.pad(infections, (len(delay) - 1, 0), constant_values=0)
    return jnp.convolve(I_padded, delay, mode="valid")


def _apply_delays(infections, delay):
    out = apply_delay(infections, delay)
    return out, out


@partial(jit, static_argnums=4)
def forward_simulate_I(m, R, gen_rev, delays, seed_L):
    infections = get_infections_intros(m, R, gen_rev, seed_L)
    infections, _ = lax.scan(_apply_delays, init=infections, xs=delays)
    return infections


v_fs_I = jit(
    vmap(forward_simulate_I, in_axes=(-1, -1, None, None, None), out_axes=-1),
    static_argnums=4,
)


@partial(jit, static_argnums=5)
def forward_simulate_I_and_prev(m, R, gen_rev, delays, inf_period, seed_L):
    infections = get_infections_intros(m, R, gen_rev, seed_L)
    delayed_infections, _ = lax.scan(_apply_delays, init=infections, xs=delays)
    prev = apply_delay(infections, inf_period)
    return delayed_infections, prev


@partial(jit, static_argnums=1)
def reporting_to_vec(rho, L):
    return jnp.pad(rho, (0, L - len(rho)), mode="wrap")


@partial(jit, static_argnums=5)
def forward_simulate_EC(I0, R, rho, gen_rev, delays, seed_L):
    infections = get_infections(I0, R, gen_rev, seed_L)
    infections, _I = lax.scan(_apply_delays, init=infections, xs=delays)
    return rho * infections[seed_L:]



from typing import Callable, List, Optional
import jax.numpy as jnp

from evofr.posterior.posterior_helpers import get_median, get_quantiles


def prep_posterior_for_plot(
    site, samples, ps: List[float], forecast: Optional[bool] = False
):
    """Prep posteriors for plotting by finding time span, medians, and quantiles.

    Parameters
    ----------

    site:
        Name of the site to access from samples.

    samples:
        Dictionary with keys being site or variable names.
        Values are Arrays containing posterior samples
        with shape (sample_number, site_shape).

    ps:
        Levels of confidence to generate quantiles for.

    forecast:
        Prep posterior for forecasts? Defaults to False.
    """
    med, quants = get_quantiles(samples, ps, site)
    t = jnp.arange(0, med.shape[0], 1)

    if forecast:
        med_f, quants_f = get_quantiles(samples, ps, site + "_forecast")
        t_f = med.shape[0] + jnp.arange(0, med_f.shape[0], 1)
        t = jnp.concatenate((t, t_f))
        med = jnp.concatenate((med, med_f))
        for i in range(len(ps)):
            quants[i] = jnp.concatenate([quants[i], quants_f[i]], axis=1)
    return t, med, quants


def plot_posterior_time(
    ax,
    t,
    med,
    quants,
    alphas: List[float],
    colors: List[str],
    included: Optional[List[bool]] = None,
):
    """
    Loop over variants to plot medians and quantiles at specifed points.
    Plots all time points unless time points to be included
    are specified in 'included'.

    Parameters
    ----------
    ax:
        Matplotlib axis to plot on.

    t:
        Time points to plot over.

    med:
        Median values.

    quants:
        Quantiles to be plotted. Organized as a list of CIs as Arrays.

    alphas:
        Transparency for each quantile.

    colors:
        List of colors to use for each variant.

    included:
        optional list of bools which determine
        which time points and variants to include observations from.
    """

    for variant in range(med.shape[-1]):
        v_included = (
            jnp.arange(0, med.shape[0])
            if included is None
            else included[:, variant]
        )
        for i in range(len(quants)):
            ax.fill_between(
                t[v_included],
                quants[i][0, v_included, variant],
                quants[i][1, v_included, variant],
                color=colors[variant],
                alpha=alphas[i],
            )
        ax.plot(t[v_included], med[v_included, variant], color=colors[variant])


def plot_R(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
    plot_neutral_line: Optional[bool] = True,
):
    t, med, quants = prep_posterior_for_plot(
        "R", samples, ps, forecast=forecast
    )
    if plot_neutral_line:
        ax.axhline(y=1.0, color="k", linestyle="--")
    plot_posterior_time(ax, t, med, quants, alphas, colors)


def plot_R_censored(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
    thres: Optional[float] = 0.001,
    plot_neutral_line: Optional[bool] = True,
):
    t, med, quants = prep_posterior_for_plot(
        "R", samples, ps, forecast=forecast
    )
    if plot_neutral_line:
        ax.axhline(y=1.0, color="k", linestyle="--")

    # Plot only variants at high enough frequency
    _, freq_median, _ = prep_posterior_for_plot(
        "freq", samples, ps, forecast=forecast
    )
    included = freq_median > thres

    plot_posterior_time(ax, t, med, quants, alphas, colors, included=included)


def plot_posterior_average_R(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    color: str,
    plot_neutral_line: Optional[bool] = True,
):
    med, V = get_quantiles(samples, ps, "R_ave")
    t = jnp.arange(0, V[-1].shape[-1], 1)

    # Make figure
    if plot_neutral_line:
        ax.axhline(y=1.0, color="k", linestyle="--")
    for i in range(len(ps)):
        ax.fill_between(
            t, V[i][0, :], V[i][1, :], color=color, alpha=alphas[i]
        )
    ax.plot(t, med, color=color)


def plot_little_r_censored(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
    thres: Optional[float] = 0.001,
    plot_neutral_line: Optional[bool] = False,
):
    t, med, quants = prep_posterior_for_plot(
        "r", samples, ps, forecast=forecast
    )
    if plot_neutral_line:
        ax.axhline(y=0.0, color="k", linestyle="--")

    # Plot only variants at high enough frequency
    _, freq_median, _ = prep_posterior_for_plot(
        "freq", samples, ps, forecast=forecast
    )
    included = freq_median > thres

    plot_posterior_time(ax, t, med, quants, alphas, colors, included=included)


def plot_posterior_frequency(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
):
    t, med, quants = prep_posterior_for_plot(
        "freq", samples, ps, forecast=forecast
    )
    plot_posterior_time(ax, t, med, quants, alphas, colors)


def plot_observed_frequency(ax, LD, colors: List[str]):
    obs_freq = jnp.divide(LD.seq_counts, LD.seq_counts.sum(axis=1)[:, None])
    N_variant = obs_freq.shape[-1]
    t = jnp.arange(0, obs_freq.shape[0])
    for variant in range(N_variant):
        ax.scatter(
            t, obs_freq[:, variant], color=colors[variant], edgecolor="black"
        )


def plot_observed_frequency_size(ax, LD, colors: List[str], size: Callable):
    N = LD.seq_counts.sum(axis=1)[:, None]
    sizes = [size(n) for n in N]
    obs_freq = jnp.divide(LD.seq_counts, LD.seq_counts.sum(axis=1)[:, None])
    N_variant = obs_freq.shape[-1]
    t = jnp.arange(0, obs_freq.shape[0])
    for variant in range(N_variant):
        ax.scatter(
            t,
            obs_freq[:, variant],
            color=colors[variant],
            s=sizes,
            edgecolor="black",
        )


def plot_posterior_I(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
):
    t, med, quants = prep_posterior_for_plot(
        "I_smooth", samples, ps, forecast=forecast
    )
    plot_posterior_time(ax, t, med, quants, alphas, colors)


def plot_posterior_smooth_EC(
    ax, samples, ps: List[float], alphas: List[float], color: str
):
    med, V = get_quantiles(samples, ps, "total_smooth_prev")
    t = jnp.arange(0, V[-1].shape[-1], 1)

    # Make figure
    for i in range(len(ps)):
        ax.fill_between(
            t, V[i][0, :], V[i][1, :], color=color, alpha=alphas[i]
        )
    ax.plot(t, med, color=color)


def plot_cases(ax, LD):
    t = jnp.arange(0, LD.cases.shape[0])
    ax.bar(t, LD.cases, color="black", alpha=0.3)


def add_dates(ax, dates, sep=1):
    t = []
    labels = []
    for (i, date) in enumerate(dates):
        if int(date.strftime("%d")) == 1:
            labels.append(date.strftime("%b"))
            t.append(i)
    ax.set_xticks(t[::sep])
    ax.set_xticklabels(labels[::sep])


def add_dates_sep(ax, dates, sep=7):
    t = []
    labels = []
    for (i, date) in enumerate(dates):
        if (i % sep) == 0:
            labels.append(date.strftime("%b %d"))
            t.append(i)
    ax.set_xticks(t)
    ax.set_xticklabels(labels)


def plot_growth_advantage(
    ax,
    samples,
    LD,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    plot_pivot_line: Optional[bool] = True,
):
    ga = jnp.array(samples["ga"])

    inds = jnp.arange(0, ga.shape[-1], 1)

    if plot_pivot_line:
        ax.axhline(y=1.0, color="k", linestyle="--")

    parts = ax.violinplot(
        ga.T, inds, showmeans=False, showmedians=False, showextrema=False
    )

    for i, pc in enumerate(parts["bodies"]):
        pc.set_facecolor(colors[i])
        pc.set_edgecolor("black")
        pc.set_alpha(1)

    q1, med, q3 = jnp.percentile(ga, jnp.array([25, 50, 75]), axis=0)
    ax.scatter(inds, med, color="white", zorder=3, edgecolor="black")
    ax.vlines(inds, q1, q3, color="k", lw=4, zorder=2)

    q1, med, q3 = jnp.percentile(ga, jnp.array([2.5, 50, 97.5]), axis=0)
    ax.vlines(inds, q1, q3, color="k", lw=2, zorder=1)

    ax.set_xticks(inds)
    ax.set_xticklabels(LD.var_names[:-1])


def plot_ga_time_censored(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
    thres: Optional[float] = 0.001,
    plot_pivot_line: Optional[bool] = True,
):
    t, med, quants = prep_posterior_for_plot(
        "ga", samples, ps, forecast=forecast
    )

    if plot_pivot_line:
        ax.axhline(y=1.0, color="k", linestyle="--")

    # Plot only variants at high enough frequency
    _, freq_median, _ = prep_posterior_for_plot(
        "freq", samples, ps, forecast=forecast
    )
    included = freq_median > thres

    plot_posterior_time(ax, t, med, quants, alphas, colors, included=included)


def plot_total_by_obs_frequency(ax, LD, total, colors: List[str]):
    T, D = LD.seq_counts.shape
    t = jnp.arange(0, T, 1)
    obs_freq = jnp.nan_to_num(
        jnp.divide(LD.seq_counts, LD.seq_counts.sum(axis=1)[:, None])
    )

    # Make figure
    bottom = jnp.zeros(t.shape)
    for variant in range(D):
        ax.bar(
            t,
            obs_freq[:, variant] * total,
            bottom=bottom,
            color=colors[variant],
        )
        bottom = obs_freq[:, variant] * total + bottom


def plot_total_by_median_frequency(ax, samples, LD, total, colors: List[str]):
    T, D = LD.seq_counts.shape
    t = jnp.arange(0, T, 1)
    med_freq = get_median(samples, "freq")

    # Make figure
    bottom = jnp.zeros(t.shape)
    for variant in range(D):
        ax.bar(
            t,
            med_freq[:, variant] * total,
            bottom=bottom,
            color=colors[variant],
        )
        bottom = med_freq[:, variant] * total + bottom


def plot_ppc_frequency(
    ax,
    samples,
    LD,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
):
    N = LD.seq_counts.sum(axis=-1)
    t, med, quants = prep_posterior_for_plot(
        "seq_counts", samples, ps, forecast=forecast
    )
    med = med / N[:, None]
    quants = [q / N[:, None] for q in quants]
    plot_posterior_time(ax, t, med, quants, alphas, colors)


def plot_ppc_seq_counts(
    ax,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
):
    t, med, quants = prep_posterior_for_plot(
        "seq_counts", samples, ps, forecast=forecast
    )
    plot_posterior_time(ax, t, med, quants, alphas, colors)


def plot_ppc_cases(
    ax, samples, ps: List[float], alphas: List[float], color: str
):
    med, V = get_quantiles(samples, ps, "cases")
    t = jnp.arange(0, V[-1].shape[-1], 1)

    # Make figure
    for i in range(len(ps)):
        ax.fill_between(
            t, V[i][0, :], V[i][1, :], color=color, alpha=alphas[i]
        )
    ax.plot(t, med, color=color)


def plot_time_varying_variant(
    ax,
    site,
    samples,
    ps: List[float],
    alphas: List[float],
    colors: List[str],
    forecast: Optional[bool] = False,
):
    t, med, quants = prep_posterior_for_plot(
        site, samples, ps, forecast=forecast
    )
    plot_posterior_time(ax, t, med, quants, alphas, colors)


def plot_time_varying_single(
    ax, site, samples, ps: List[float], alphas: List[float], color: str
):
    med, V = get_quantiles(samples, ps, site)
    t = jnp.arange(0, V[-1].shape[-1], 1)

    # Make figure
    for i in range(len(ps)):
        ax.fill_between(
            t, V[i][0, :], V[i][1, :], color=color, alpha=alphas[i]
        )
    ax.plot(t, med, color=color)



from .data_helpers import prep_dates, prep_cases
from .data_spec import DataSpec
import pandas as pd
from typing import Optional


class CaseCounts(DataSpec):
    def __init__(
        self, raw_cases: pd.DataFrame, date_to_index: Optional[dict] = None
    ):
        """Construct a data specification for handling case counts.

        Parameters
        ----------
        raw_cases:
            a dataframe containing case counts with columns 'cases' and 'date'.

        date_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        Returns
        -------
        CaseCounts
        """

        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_cases["date"])
        self.date_to_index = date_to_index

        self.cases = prep_cases(raw_cases, self.date_to_index)

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()
        data["cases"] = self.cases
        return data



from typing import List, Optional
import numpy as np
import pandas as pd

from evofr.data.data_spec import DataSpec
from evofr.data.data_helpers import prep_dates, prep_sequence_counts
from evofr.models.renewal_model.model_options import MultinomialSeq
from .multinomial_logistic_regression import MultinomialLogisticRegression
import jax.numpy as jnp

import numpyro
import numpyro.distributions as dist

from .model_spec import ModelSpec

from functools import partial


def mutational_fitness_model(
    seq_counts,
    N,
    X,
    mutation_presence,
    SeqLik,
    tau=None,
    pred=False,
):
    T, N_variants = seq_counts.shape
    _, N_features = X.shape
    _, N_mutations = mutation_presence.shape

    # Sampling parameters for growth advantages
    # Sampling intercepts
    raw_alpha = (
        numpyro.sample(
            "raw_alpha",
            dist.Normal(0.0, 1.0),
            sample_shape=(1, N_variants - 1),
        )
        * 10.0
    )
    raw_alpha = jnp.append(raw_alpha, 0.0)

    # Sampling mutation innovations
    raw_delta = numpyro.sample(
        "raw_delta", dist.Normal(0.0, 1.0), sample_shape=(N_mutations,)
    )
    # Innovations to beta
    raw_beta = jnp.dot(mutation_presence, raw_delta)
    raw_beta = numpyro.deterministic("raw_beta", raw_beta - raw_beta[-1])

    # How do we ensure identifiablity?
    # We need to make sure everything is relative to pivot
    # But we won't be able to estimate mutational fitness of things which ...
    # We can have effect by lineage for re-occuring mutations :-D.

    beta = numpyro.deterministic(
        "beta", jnp.row_stack((raw_alpha, raw_beta))
    )  # All parameters are relative to last column / variant

    logits = jnp.dot(X, beta)  # Logit frequencies by variant

    # Compute frequency
    _freq = jnp.exp(logits)
    freq = numpyro.deterministic(
        "freq", jnp.clip(_freq / _freq.sum(axis=-1)[:, None], 1e-12, 1 - 1e-12)
    )

    # Evaluate likelihood of sequence counts given delays
    SeqLik.model(seq_counts, N, freq, pred)  # Evaluate likelihood

    # Compute growth advantage from model
    if tau is not None:
        numpyro.deterministic("ga", jnp.exp(raw_beta * tau)[:-1])


class MutationalFitnessMLR(ModelSpec):
    def __init__(
        self,
        tau: float,
        SeqLik=None,
    ) -> None:
        """Construct ModelSpec for MLR allowing derived variants
        to have mutation derived fitnesses.

        Parameters
        ----------
        tau:
            Assumed generation time for conversion to relative R.

        SeqLik:
            Optional sequence likelihood option: MultinomialSeq or
            DirMultinomialSeq. Defaults to MultinomialSeq.

        Returns
        -------
        MutationalFitnessMLR
        """
        self.tau = tau  # Fixed generation time
        self.SeqLik = MultinomialSeq() if SeqLik is None else SeqLik
        self.model_fn = partial(
            mutational_fitness_model,
            SeqLik=self.SeqLik,
            tau=self.tau,
        )

    def augment_data(self, data: dict) -> None:
        T = len(data["N"])
        data["X"] = MultinomialLogisticRegression.make_ols_feature(
            0, T
        )  # Use intercept and time as predictors


def prep_mutations(raw_mutations, var_names):
    # Fitler to only variants in var_names
    rw = raw_mutations[raw_mutations.variant.isin(var_names)]

    # Find mutations of interest
    mutations = rw["mutation"].unique()
    num_variants, num_muts = len(var_names), mutations.size
    mut_matrix = np.zeros((num_variants, num_muts))
    for v, variant in enumerate(var_names):
        v_muts = rw[rw.variant == variant]["mutation"].values
        mut_matrix[v, :] = np.isin(mutations, v_muts)

    # Drop rows that are present in everything...
    is_fixed = mut_matrix.sum(axis=0) == num_variants
    return mutations[~is_fixed], mut_matrix[:, ~is_fixed]


class MutationalFitnessSequenceCounts(DataSpec):
    def __init__(
        self,
        raw_seq: pd.DataFrame,
        raw_mutations: pd.DataFrame,
        date_to_index: Optional[dict] = None,
        var_names: Optional[List] = None,
        pivot: Optional[str] = None,
    ):
        """Construct a data specification for handling variant frequencies
        for mutational fitness models.

         Parameters
         ----------
         raw_seq:
             a dataframe containing sequence counts with columns 'sequences',
             'variant', and date'.

         raw_mutations:
             a dataframe containing variants and their mutations with columns
             'variant' and 'mutatation'.

         date_to_index:
             optional dictionary for mapping calender dates to nd.array indices

         var_names:
             optional list containing names of variants to be present

         Returns
         -------
        MutationalFitnessSequenceCounts
        """

        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_seq["date"])
        self.date_to_index = date_to_index

        # Turn dataframe to counts of each variant sequenced each day
        self.var_names, self.seq_counts = prep_sequence_counts(
            raw_seq, self.date_to_index, var_names, pivot=pivot
        )
        self.pivot = self.var_names[-1]

        # Create mutation indicator matrix
        self.mut_names, self.mutation_presence = prep_mutations(
            raw_mutations, self.var_names
        )

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()

        # Export `seq_counts`, sequences each day, and variant names
        data["seq_counts"] = self.seq_counts
        data["N"] = self.seq_counts.sum(axis=-1)
        data["mutation_presence"] = self.mutation_presence
        return data



import jax.numpy as jnp
import numpy as np
import numpyro
import numpyro.distributions as dist

from .LAS import LaplaceRandomWalk
from .model_helpers import is_obs_idx, pad_to_obs


class FixedGA:
    def __init__(self, gam_prior=0.5, prior_family="Cauchy"):
        """Construct FixedGrowthAdvantage model.
        Parameters
        ----------
        gam_prior:
            hyper_prior for scale of Rt increment.
        prior_family:
            family to draw scale from for LAS.
        Returns
        -------
        FixedGA
        """
        self.gam_prior = gam_prior

        if prior_family == "Cauchy":
            self.scale_prior = dist.HalfCauchy()
        elif prior_family == "Normal":
            self.scale_prior = dist.HalfNormal()

    def model(self, N_variant, X):
        _, k = X.shape

        # Locally adaptive smoothing
        gam = numpyro.sample("gam", self.scale_prior) * self.gam_prior
        beta_0 = numpyro.sample("beta_0", dist.Normal(0.0, 1.0))
        with numpyro.plate("N_steps_base", k - 1):
            beta_rw_step = numpyro.sample("beta_rw_step", dist.Laplace()) * gam
            beta_rw = numpyro.deterministic(
                "beta_rw", jnp.cumsum(beta_rw_step)
            )

        # Combine increments and starting position
        beta_rw = jnp.append(jnp.zeros(1), beta_rw)
        beta = beta_0 + beta_rw

        # Getting log variant growth advantages
        with numpyro.plate("N_variant_m1", N_variant - 1):
            v = numpyro.sample("v", dist.Normal(0.0, 0.5))

        numpyro.deterministic(
            "ga", jnp.exp(v)
        )  # Transform to growth advantage

        # Computing R
        R = numpyro.deterministic(
            "R", jnp.exp((X @ beta + jnp.append(v, 0.0)[:, None])).T
        )
        return R


class FreeGrowth:
    def __init__(self, gam_prior=0.5, prior_family="Cauchy"):
        self.gam_prior = gam_prior

        if prior_family == "Cauchy":
            self.scale_prior = dist.HalfCauchy()
        elif prior_family == "Normal":
            self.scale_prior = dist.HalfNormal()

    def model(self, N_variant, X):
        _, k = X.shape

        # Locally adaptive smoothing on all R trajectories
        with numpyro.plate("variant_beta", N_variant):
            gam = numpyro.sample("gam", self.scale_prior) * self.gam_prior
            beta_0 = numpyro.sample("beta_0", dist.Normal(0.0, 0.5))
            with numpyro.plate("N_steps_base", k - 1):
                beta_rw_step = (
                    numpyro.sample("beta_rw_step", dist.Laplace()) * gam
                )
                beta_rw = numpyro.deterministic(
                    "beta_rw", jnp.cumsum(beta_rw_step, axis=0)
                )

        beta_rw = jnp.row_stack((jnp.zeros((1, N_variant)), beta_rw))
        beta = beta_0 + beta_rw

        # Computing R
        R = numpyro.deterministic("R", jnp.exp(X @ beta))
        return R


class GARW:
    def __init__(
        self, gam_prior=0.5, gam_delta_prior=0.5, prior_family="Cauchy"
    ):
        """Construct GARW model.
        Parameters
        ----------
        gam_prior:
            hyper_prior for scale of Rt increment.
        gam_delta_prior:
            hyper_prior for scale of growth advantage increment.
        prior_family:
            family to draw scale from for LAS.

        Returns
        -------
        GARW
        """

        self.gam_prior = gam_prior
        self.gam_delta_prior = gam_delta_prior

        if prior_family == "Cauchy":
            self.scale_prior = dist.HalfCauchy()
        elif prior_family == "Normal":
            self.scale_prior = dist.HalfNormal()

    def model(self, N_variant, X):
        _, k = X.shape

        # Time varying base trajectory
        beta_0 = numpyro.sample("beta_0", dist.Normal(0.0, 1.0)) * 0.2

        # Simulate increments for base trajectory
        gam = numpyro.sample("gam", self.scale_prior) * self.gam_prior
        # beta_rw = numpyro.sample(
        #     "beta_rw", LaplaceRandomWalk(scale=gam, num_steps=k - 1)
        # )
        with numpyro.plate("N_steps_base", k - 1):
            beta_rw_step = numpyro.sample("beta_rw_step", dist.Laplace()) * gam
            beta_rw = numpyro.deterministic(
                "beta_rw", jnp.cumsum(beta_rw_step)
            )

        # Combine increments and starting position
        beta_rw = jnp.append(jnp.zeros(1), beta_rw)
        beta = beta_0 + beta_rw

        # Time varying growth advantage as random walk
        # Regularizes changes in growth advantage of variants
        with numpyro.plate("N_variant_m1", N_variant - 1):
            delta_0 = numpyro.sample("delta_0", dist.Normal(0.0, 1.0)) * 0.2
            gam_delta = (
                numpyro.sample("gam_delta", self.scale_prior)
                * self.gam_delta_prior
            )
            # delta_rw = numpyro.sample(
            #     "delta_rw", LaplaceRandomWalk(scale=gam_delta, num_steps=k - 1)
            # )

            with numpyro.plate("N_steps_variant", k - 1):
                delta_rw_step = (
                    numpyro.sample("delta_rw_step", dist.Laplace()) * gam_delta
                )
                delta_rw = numpyro.deterministic(
                    "delta_rw", jnp.cumsum(delta_rw_step, axis=0)
                )

            delta_rw = jnp.concatenate(
                (jnp.zeros((1, N_variant - 1)), delta_rw), axis=0
            )
            delta = delta_0 + delta_rw

        # Transform to growth advatnage
        numpyro.deterministic("ga", jnp.exp(X @ delta))

        # Construct beta matrix
        beta_mat = beta[:, None] + jnp.hstack((delta, jnp.zeros((k, 1))))
        R = numpyro.deterministic("R", jnp.exp((X @ beta_mat)))
        return R


class GAPRW:
    def __init__(self, gam_prior=0.5, gam_delta_prior=0.5):
        self.gam_prior = gam_prior
        self.gam_delta_prior = gam_delta_prior

    def model(self, N_variant, X):
        T, _ = X.shape

        # Time varying base trajectory
        gam = numpyro.sample("gam", dist.HalfNormal()) * self.gam_prior
        beta_0 = numpyro.sample("beta_0", dist.Normal(0.0, 1.0))
        with numpyro.plate("N_steps_base", T - 1):
            beta_rw_step = numpyro.sample("beta_rw_step", dist.Laplace()) * gam
            beta_rw = numpyro.deterministic(
                "beta_rw", jnp.cumsum(beta_rw_step)
            )

        # Combine increments and starting position
        beta_rw = jnp.append(jnp.zeros(1), beta_rw)
        beta = beta_0 + beta_rw

        # Time varying growth advantage as random walk
        # Regularizes changes in growth advantage of variants
        with numpyro.plate("N_variant_m1", N_variant - 1):
            delta_0 = numpyro.sample("delta_0", dist.Normal(0.0, 0.5))
            gam_delta = numpyro.sample(
                "gam_delta", dist.HalfNormal(self.gam_delta_prior)
            )
            # delta_rw = numpyro.sample(
            #     "delta_rw", LaplaceRandomWalk(scale=gam_delta, num_steps=T - 1)
            # )
            # delta_rw = jnp.concatenate(
            #     (jnp.zeros((N_variant - 1, 1)), delta_rw), axis=-1
            # )
            #
            # delta = delta_0 + delta_rw.T
            with numpyro.plate("N_steps_variant", T - 1):
                delta_rw_step = (
                    numpyro.sample("delta_rw_step", dist.Laplace()) * gam_delta
                )
                delta_rw = numpyro.deterministic(
                    "delta_rw", jnp.cumsum(delta_rw_step, axis=0)
                )

            delta_rw = jnp.concatenate(
                (jnp.zeros((1, N_variant - 1)), delta_rw), axis=0
            )
            delta = delta_0 + delta_rw

        # Transform to growth advatnage
        numpyro.deterministic("ga", jnp.exp(delta))

        # Construct beta matrix
        beta_mat = beta[:, None] + jnp.hstack((delta, jnp.zeros((T, 1))))
        R = numpyro.deterministic("R", jnp.exp(beta_mat))
        return R


class PoisCases:
    def __init__(self):
        pass

    def model(self, cases, EC, pred=False):
        is_obs = is_obs_idx(cases)  # Find unobserved case counts
        # Overwrite defaults for predictive checks
        obs = None if pred else np.nan_to_num(cases)

        numpyro.sample("cases", dist.Poisson(pad_to_obs(EC, is_obs)), obs=obs)


class ZIPoisCases:
    def __init__(self):
        pass

    def model(self, cases, EC, pred=False):

        # Finding zero locations and making only parameters for zero observation
        is_zero = cases == 0
        zero_idx = np.nonzero(is_zero)[0]

        gate = jnp.zeros_like(cases) + 1e-12
        if zero_idx.shape[0] > 0:
            with numpyro.plate("n_zero", zero_idx.shape[0]):
                zp = numpyro.sample("zp", dist.Beta(0.1, 0.1))
            gate = gate.at[zero_idx].set(zp)

        # Overwrite defaults for predictive checks
        gate = jnp.zeros_like(cases) if pred else gate
        obs = None if pred else np.nan_to_num(cases)

        numpyro.sample(
            "cases", dist.ZeroInflatedPoisson(rate=EC, gate=gate), obs=obs
        )


class NegBinomCases:
    def __init__(self, raw_alpha_sd=0.01):
        self.raw_alpha_sd = raw_alpha_sd

    def model(self, cases, EC, pred=False):
        # NegativeBinomial sampling
        raw_alpha = numpyro.sample(
            "raw_alpha", dist.HalfNormal(self.raw_alpha_sd)
        )
        is_obs = is_obs_idx(cases)  # Find unobserved case counts

        # Overwrite defaults for predictive checks
        obs = None if pred else np.nan_to_num(cases)

        numpyro.sample(
            "cases",
            dist.NegativeBinomial2(
                mean=pad_to_obs(EC, is_obs),
                concentration=jnp.power(raw_alpha, -2),
            ),
            obs=obs,
        )


class ZINegBinomCases:
    def __init__(self, raw_alpha_sd=0.01):
        self.raw_alpha_sd = raw_alpha_sd

    def model(self, cases, EC, pred=False):
        # NegativeBinomial sampling
        raw_alpha = (
            numpyro.sample("raw_alpha", dist.HalfNormal()) * self.raw_alpha_sd
        )

        # Finding zero locations and making only parameters for zero observation
        is_zero = cases == 0
        zero_idx = np.nonzero(is_zero)[0]

        gate = jnp.zeros_like(cases) + 1e-12
        if zero_idx.shape[0] > 0:
            with numpyro.plate("n_zero", zero_idx.shape[0]):
                zp = numpyro.sample("zp", dist.Beta(0.1, 0.1))
            gate = gate.at[zero_idx].set(zp)

        # Overwrite defaults for predictive checks
        gate = jnp.zeros_like(cases) if pred else gate
        obs = None if pred else np.nan_to_num(cases)

        numpyro.sample(
            "cases",
            dist.ZeroInflatedNegativeBinomial2(
                mean=EC, gate=gate, concentration=jnp.power(raw_alpha, -2)
            ),
            obs=obs,
        )


class MultinomialSeq:
    def __init__(self):
        pass

    def model(self, seq_counts, N, freq, pred=False):

        # Sample with Multinomial
        obs = None if pred else np.nan_to_num(seq_counts)
        numpyro.sample(
            "seq_counts",
            dist.Multinomial(total_count=np.nan_to_num(N), probs=freq),
            obs=obs,
        )


class DirMultinomialSeq:
    def __init__(self, xi_prior=None):
        self.xi_prior = xi_prior if xi_prior is not None else 99

    def model(self, seq_counts, N, freq, pred=False):
        # Overdispersion in sequence counts
        xi = numpyro.sample("xi", dist.Beta(1, self.xi_prior))
        trans_xi = jnp.reciprocal(xi) - 1
        # xi = numpyro.sample("xi", dist.HalfNormal())
        # trans_xi = self.xi_prior / xi

        obs = None if pred else np.nan_to_num(seq_counts)

        # Sample with DirichletMultinomial
        numpyro.sample(
            "seq_counts",
            dist.DirichletMultinomial(
                total_count=np.nan_to_num(N),
                concentration=1e-8 + trans_xi * freq,
            ),
            obs=obs,
        )



from typing import Optional
from jax import vmap, Array
import jax.numpy as jnp
from jax.scipy.special import gammaln
from .basis_fns import BasisFunction


class HSGaussianProcess(BasisFunction):
    """
    Implementation of basis approximation to Gaussian processes.
    """

    def __init__(self):
        pass

    @staticmethod
    def lam(L: float, m: int):
        return jnp.square(m * jnp.pi / (2 * L))

    @staticmethod
    def phi(L: float, m: int, x: float):
        lam = HSGaussianProcess.lam(L, m)
        arg = jnp.sqrt(lam) * (x + L)
        return jnp.sqrt(1 / L) * jnp.sin(arg)

    @staticmethod
    def phi_matrix(L: float, m: int, x: float):
        phi_mapped = vmap(
            HSGaussianProcess.phi, in_axes=(None, 0, None), out_axes=-1
        )
        return phi_mapped(L, m, x)


class SquaredExponential(HSGaussianProcess):
    def __init__(
        self,
        alpha: Optional[float] = None,
        rho: Optional[float] = None,
        L: Optional[float] = None,
        m: Optional[int] = None,
    ):
        self.alpha = alpha if alpha else 1.0
        self.rho = rho if rho else 1.0
        self.L = L if L else 10.0
        self.m = m if m else 5

    @staticmethod
    def spd(alpha: float, rho: float, w: float):
        return (
            alpha
            * jnp.sqrt(2 * jnp.pi)
            * rho
            * jnp.exp(-0.5 * jnp.square(rho) * jnp.square(w))
        )

    def make_features(self, data: dict) -> Array:
        T = data["N"].shape[0]

        # Make time period
        ts = jnp.arange(T)

        # Make eigenvalues
        ms = jnp.arange(1, self.m + 1)
        lams = self.lam(self.L, ms)

        # Make eigenvectors
        phi = self.phi_matrix(self.L, ms, ts)

        # Make spectral density
        delta = self.spd(self.alpha, self.rho, jnp.sqrt(lams))

        # Return feature matrix
        return phi * jnp.sqrt(delta)


class Matern(HSGaussianProcess):
    def __init__(
        self,
        alpha: Optional[float] = None,
        rho: Optional[float] = None,
        nu: Optional[float] = None,
        L: Optional[float] = None,
        m: Optional[int] = None,
    ):
        self.alpha = alpha if alpha else 1.0
        self.rho = rho if rho else 1.0
        self.nu = nu if nu else 1.5
        self.L = L if L else 10.0
        self.m = m if m else 5

    @staticmethod
    def spd(alpha: float, rho: float, nu: float, w: float):
        gammanu = jnp.exp(gammaln(nu))
        gammanuhalf = jnp.exp(gammaln(nu + 0.5))
        coef = (
            alpha
            * 2
            * jnp.sqrt(jnp.pi)
            * gammanuhalf
            * jnp.power(2 * nu, nu)
            / (gammanu * jnp.power(rho, 2 * nu))
        )
        base = 2 * nu * jnp.power(rho, -2) + 4 * jnp.square(jnp.pi * w)
        expon = -nu - 0.5
        return coef * jnp.power(base, expon)

    def make_features(self, data: dict) -> Array:
        T = data["N"].shape[0]

        # Make time period
        ts = jnp.arange(T)

        # Make eigenvalues
        ms = jnp.arange(1, self.m + 1)
        lams = self.lam(self.L, ms)

        # Make eigenvectors
        phi = self.phi_matrix(self.L, ms, ts)

        # Make spectral density
        delta = self.spd(self.alpha, self.rho, self.nu, jnp.sqrt(lams))

        # Return feature matrix
        return phi * jnp.sqrt(delta)



import jax.numpy as jnp

from evofr.models.model_spec import ModelSpec

from .model_factories import _spline_incidence_model_factory
from .basis_functions import Spline, SplineDeriv


class SplineIncidenceModel(ModelSpec):
    def __init__(
        self,
        k=None,
        CLik=None,
        SLik=None,
    ):
        self.k = k if k is not None else 20
        self.CLik = CLik
        self.SLik = SLik
        self.make_model()

    def make_model(self):
        self.model_fn = _spline_incidence_model_factory(
            self.CLik,
            self.SLik,
        )

    def augment_data(self, data, order=4):
        T = len(data["cases"])
        s = jnp.linspace(0, T, self.k)
        data["X"] = Spline.matrix(jnp.arange(T), s, order=order)
        data["X_prime"] = SplineDeriv.matrix(jnp.arange(T), s, order=order)



from typing import Optional
from jax import vmap, Array
import jax.numpy as jnp
from .basis_fns import BasisFunction


class Spline(BasisFunction):
    def __init__(
        self,
        s: Optional[Array] = None,
        order: Optional[int] = None,
        k: Optional[int] = None,
    ):
        """Construct Spline class.

        Parameters
        ----------
        s:
            Optional knot points for spline basis.
            Defaults to 'k' evenly spaced points if absent.

        order:
            Optional order for splines.
            Defaults to 4 corresponding to cubic splines.

        k:
            Optional number of basis functions.
            Must pass either 's' or 'k'.

        Returns
        -------
        Spline
        """
        self.s = s
        self.order = order if order else 4
        self.k = k  # Need to handle error if neither s and k are passed

    @staticmethod
    def _omega(s1, s2, t):
        return jnp.where(s1 == s2, jnp.zeros_like(t), (t - s1) / (s2 - s1))

    @staticmethod
    def _basis(t, s, order, i):
        if order == 1:
            return jnp.where(
                (t >= s[i]) * (t < s[i + 1]),
                jnp.ones_like(t),
                jnp.zeros_like(t),
            )

        # Recurse left
        w1 = Spline._omega(s[i], s[i + order - 1], t)
        B1 = Spline._basis(t, s, order - 1, i)

        # Recurse right
        w2 = Spline._omega(s[i + 1], s[i + order], t)
        B2 = Spline._basis(t, s, order - 1, i + 1)
        return w1 * B1 + (1 - w2) * B2

    @staticmethod
    def matrix(t, s, order):
        """Construct matrix for spline of
        order 'order' with knots 's' at points 't'.
        """
        _s = jnp.pad(s, mode="edge", pad_width=(order - 1))  # Extend knots
        X = vmap(lambda i: Spline._basis(t, _s, order, i))(
            jnp.arange(0, len(s) + order - 2)
        )  # Make spline basis
        return X.T

    def make_features(
        self, data: Optional[dict] = None, T: Optional[float] = None
    ) -> Array:
        # Check for maximum time
        if T is None and data is not None:
            T = data["N"].shape[0]

        # If pivots not defined, make self.k equally spaced splines
        if self.s is None and self.k:
            self.s = jnp.linspace(0, T, self.k)

        return self.matrix(jnp.arange(T), self.s, self.order)


class SplineDeriv:
    def __init__(
        self,
        s: Optional[Array] = None,
        order: Optional[int] = None,
        k: Optional[int] = None,
    ):
        """Construct SplineDeriv class. Represents the derivative of the Spline class.

        Parameters
        ----------
        s:
            Optional knot points for spline basis.
            Defaults to 'k' evenly spaced points if absent.

        order:
            Optional order for splines.
            Defaults to 4 corresponding to cubic splines.

        k:
            Optional number of basis functions.
            Must pass either 's' or 'k'.

        Returns
        -------
        SplineDeriv
        """

        self.s = s
        self.order = order if order else 4
        self.k = k  # Need to handle error if neither s and k are passed

    @staticmethod
    def _omegap(s1, s2, t):
        return jnp.where(s1 == s2, jnp.zeros_like(t), jnp.reciprocal(s2 - s1))

    @staticmethod
    def _basis(t, s, order, i):
        if order == 1:
            return jnp.where(
                (t >= s[i]) * (t < s[i + 1]),
                jnp.ones_like(t),
                jnp.zeros_like(t),
            )

        # Recurse left
        w1 = SplineDeriv._omegap(s[i], s[i + order - 1], t)
        B1 = Spline._basis(t, s, order - 1, i)

        # Recurse right
        w2 = SplineDeriv._omegap(s[i + 1], s[i + order], t)
        B2 = Spline._basis(t, s, order - 1, i + 1)
        return (order - 1) * (w1 * B1 - w2 * B2)

    @staticmethod
    def matrix(t, s, order):
        """Construct matrix for spline derivative of
        order 'order' with knots 's' at points 't'.
        """

        _s = jnp.pad(s, mode="edge", pad_width=(order - 1))  # Extend knots
        X = vmap(lambda i: SplineDeriv._basis(t, _s, order, i))(
            jnp.arange(0, len(s) + order - 2)
        )  # Make spline basis
        return X.T

    def make_features(
        self, data: Optional[dict] = None, T: Optional[float] = None
    ) -> Array:
        # Check for maximum time
        if T is None and data is not None:
            T = data["N"].shape[0]

        # If pivots not defined, make self.k equally spaced splines
        if self.s is None and self.k:
            self.s = jnp.linspace(0, T, self.k)

        return self.matrix(jnp.arange(T), self.s, self.order)



# from typing import Dict
# from evofr.models.model_spec import ModelSpec
#
#
# class BlackJaxPyMc:
#     @staticmethod
#     def init(key, model: ModelSpec, data: Dict):
#         if not hasattr(model, "pymc_model"):
#             return None, lambda _: None
#         pymc_model = model.pymc_model
#         logdensity_fn = get_jaxified_logp(pymc_model)
#         rvs = [rv.name for rv in pymc_model.value_vars]
#         init_position_dict = pymc_model.initial_point()
#         init_position = [init_position_dict[rv] for rv in rvs]
#         return init_position, logdensity_fn
#
#
# class BlackJaxAesara:
#     @staticmethod
#     def init(key, model: ModelSpec, data: Dict):
#         if hasattr(model, "logprob"):
#
#             def logdensity_fn(position):
#                 flat_position = tuple(position.values())
#                 return model.logprob(*flat_position, **data)
#
#         else:
#
#             def logdensity_fn(position):
#                 return None
#
#         if hasattr(model, "initial_position_fn"):
#             init_position = model.initial_position_fn(key)
#         else:
#             init_position = None
#         return init_position, logdensity_fn



import jax.numpy as jnp
import jax.random as random

import numpyro
from numpyro.distributions import constraints, HalfCauchy, Laplace
from numpyro.distributions.distribution import Distribution
from numpyro.distributions.util import (
    is_prng_key,
    validate_sample,
)


class LaplaceRandomWalk(Distribution):
    """Laplace random walk based on numpyro built in Gaussian Random Walk."""
    arg_constraints = {"scale": constraints.positive}
    support = constraints.real_vector
    reparametrized_params = ["scale"]

    def __init__(self, scale=1.0, num_steps=1, validate_args=None):
        assert (
            isinstance(num_steps, int) and num_steps > 0
        ), "`num_steps` argument should be an positive integer."
        self.scale = scale
        self.num_steps = num_steps
        batch_shape, event_shape = jnp.shape(scale), (num_steps,)
        super(LaplaceRandomWalk, self).__init__(
            batch_shape, event_shape, validate_args=validate_args
        )

    def sample(self, key, sample_shape=()):
        assert is_prng_key(key)
        shape = sample_shape + self.batch_shape + self.event_shape
        walks = random.laplace(key, shape=shape)
        return jnp.cumsum(walks, axis=-1) * jnp.expand_dims(self.scale, axis=-1)

    @validate_sample
    def log_prob(self, value):
        init_prob = Laplace(0.0, self.scale).log_prob(value[..., 0])
        scale = jnp.expand_dims(self.scale, -1)
        step_probs = Laplace(value[..., :-1], scale).log_prob(value[..., 1:])
        return init_prob + jnp.sum(step_probs, axis=-1)

    @property
    def mean(self):
        return jnp.zeros(self.batch_shape + self.event_shape)

    @property
    def variance(self):
        return jnp.broadcast_to(
            jnp.expand_dims(self.scale, -1) ** 2 * jnp.arange(1, self.num_steps + 1),
            self.batch_shape + self.event_shape,
        )

    def tree_flatten(self):
        return (self.scale,), self.num_steps

    @classmethod
    def tree_unflatten(cls, aux_data, params):
        return cls(*params, num_steps=aux_data)


def LAS_Laplace(beta_name, k):
    gam = numpyro.sample("gam", HalfCauchy(0.5))
    beta = numpyro.sample(beta_name, LaplaceRandomWalk(scale=gam, num_steps=k))
    return beta



from functools import partial
from typing import Optional
import numpy as np
import jax.numpy as jnp
from jax import vmap
from jax.nn import softmax

import numpyro
import numpyro.distributions as dist

from .renewal_model.model_options import DirMultinomialSeq
from .model_spec import ModelSpec


def simulate_MLR_freq(growth_advantage, freq0, tau, max_time):
    times = np.arange(max_time)
    delta = np.log(growth_advantage) / tau  # to relative fitness
    ufreq = freq0 * np.exp(delta * times[..., None])
    return ufreq / ufreq.sum(axis=-1)[..., None]


def simulate_MLR(growth_advantage, freq0, tau, Ns):
    max_time = len(Ns)
    freq = simulate_MLR_freq(growth_advantage, freq0, tau, max_time)
    seq_counts = [
        np.random.multinomial(Ns[t], freq[t, :]) for t in range(max_time)
    ]
    return freq, np.stack(seq_counts)


def MLR_numpyro(
    seq_counts,
    N,
    X,
    tau=None,
    pred=False,
    var_names=None,
    dir_multinomial=False,
    xi_prior=None,
):
    _, N_variants = seq_counts.shape
    _, N_features = X.shape

    # Sampling parameters
    raw_beta = numpyro.sample(
        "raw_beta",
        dist.Normal(0.0, 3.0),
        sample_shape=(N_features, N_variants - 1),
    )

    beta = numpyro.deterministic(
        "beta",
        jnp.column_stack(
            (raw_beta, jnp.zeros(N_features))
        ),  # All parameters are relative to last column / variant
    )

    logits = jnp.dot(X, beta)  # Logit frequencies by variant
    freq = numpyro.deterministic("freq", softmax(logits, axis=-1))

    # Evaluate likelihood using either Multinomial or Dirichlet-multinomial
    if dir_multinomial:
        seq_likelihood = DirMultinomialSeq(xi_prior=xi_prior)
        seq_likelihood.model(seq_counts, N, freq, pred=pred)
    else:
        obs = None if pred else np.nan_to_num(seq_counts)
        numpyro.sample(
            "seq_counts",
            dist.MultinomialLogits(
                logits=logits, total_count=np.nan_to_num(N)
            ),
            obs=obs,
        )

    # Compute growth advantage from model
    if tau is not None:
        numpyro.deterministic(
            "ga", jnp.exp(beta[-1, :-1] * tau)
        )  # Last row corresponds to linear predictor / growth advantage


class MultinomialLogisticRegression(ModelSpec):
    def __init__(
        self,
        tau: float,
        dir_multinomial=False,
        xi_prior: Optional[float] = None,
    ) -> None:
        """Construct ModelSpec for Multinomial logistic regression

        Parameters
        ----------
        tau:
            Assumed generation time for conversion to relative R.
        dir_multinomial:
            Whether to use a Dirichlet-Multinomial likelihood for sequence counts.
            Default is False.
        xi_prior:
            Prior on seq_counts overdispersion magnitude only used when dir_multinomial == True.

        Returns
        -------
        MultinomialLogisticRegression
        """
        self.tau = tau  # Fixed generation time
        self.dir_multinomial = dir_multinomial
        self.xi_prior = xi_prior
        self.model_fn = partial(
            MLR_numpyro, dir_multinomial=self.dir_multinomial, xi_prior=self.xi_prior
        )

    @staticmethod
    def make_ols_feature(start, stop):
        """
        Construct simple OLS features (1, x) for MultinomialLogisticRegression.

        Parameters
        ----------
        start:
            Start value for OLS feature.
        stop:
            Stop value for OLS feature.
        """
        t = jnp.arange(start=start, stop=stop)
        return jnp.column_stack((jnp.ones_like(t), t))

    def augment_data(self, data: dict) -> None:
        T = len(data["N"])
        data["tau"] = self.tau
        data["X"] = self.make_ols_feature(
            0, T
        )  # Use intercept and time as predictors

    @staticmethod
    def forecast_frequencies(samples, forecast_L):
        """
        Use posterior beta to forecast posterior frequenicies.
        """

        # Making feature matrix for forecasting
        last_T = samples["freq"].shape[1]
        X = MultinomialLogisticRegression.make_ols_feature(
            start=last_T, stop=last_T + forecast_L
        )

        # Posterior beta
        beta = jnp.array(samples["beta"])

        # Matrix multiplication by sample
        dot_by_sample = vmap(jnp.dot, in_axes=(None, 0), out_axes=0)
        logits = dot_by_sample(X, beta)  # Logit frequencies by variant
        samples["freq_forecast"] = softmax(logits, axis=-1)
        return samples



from .variant_frequencies import VariantFrequencies
from .data_helpers import prep_dates, format_var_names
from .data_spec import DataSpec
import numpy as np
import pandas as pd
from typing import Optional


class HierFrequencies(DataSpec):
    def __init__(
        self,
        raw_seq: pd.DataFrame,
        group: str,
        date_to_index: Optional[dict] = None,
        pivot: Optional[str] = None,
    ):
        """Construct a data specification for handling variant frequencies
        in hierarchical models.

        Parameters
        ----------
        raw_seq:
            a dataframe containing sequence counts with columns 'sequences',
            'variant', and date'.

        group:
            string defining which column to seperate data by.

        date_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        pivot:
            optional name of variant to place last.
            Defaults to "other" if present otherwise.
            This will usually used as a reference or pivot strain.

        Returns
        -------
        HierFrequencies
        """
        # Get date to index
        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_seq["date"])
        self.date_to_index = date_to_index

        # Get variant names
        raw_var_names = list(pd.unique(raw_seq.variant))
        raw_var_names.sort()
        self.var_names = format_var_names(raw_var_names, pivot=pivot)
        self.pivot = self.var_names[-1]

        # Loop each group
        grouped = raw_seq.groupby(group)
        self.names = [name for name, _ in grouped]
        self.groups = [
            VariantFrequencies(group, self.date_to_index, self.var_names)
            for _, group in grouped
        ]

        self.seq_counts = np.stack(
            [g.seq_counts for g in self.groups], axis=-1
        )

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()
        data["seq_counts"] = np.stack(
            [g.seq_counts for g in self.groups], axis=-1
        )
        data["N"] = np.stack(
            [g.seq_counts.sum(axis=-1) for g in self.groups], axis=-1
        )
        data["var_names"] = self.var_names
        return data



import datetime
import numpy as np
import pandas as pd
from typing import List, Optional
import warnings


def prep_dates(raw_dates: pd.Series):
    """Return vector of dates and a mapping of dates to indices.

    Parameters
    ----------
    raw_dates:
        pandas series containing dates of interest

    Returns
    -------
    dates:
        list containing dates

    date_to_index:
        dictionary taking in dates and returning integer indices
    """
    _dates = pd.to_datetime(raw_dates)
    dmn = _dates.min()
    dmx = _dates.max()
    n_days = (dmx - dmn).days
    dates = [dmn + datetime.timedelta(days=d) for d in range(0, 1 + n_days)]
    date_to_index = {d: i for (i, d) in enumerate(dates)}
    return dates, date_to_index


def forecast_dates(dates, T_forecast: int):
    """Generate dates of forecast given forecast interval of length 'T_forecast'."""
    last_date = dates[-1]
    dates_f = []
    for d in range(T_forecast):
        dates_f.append(last_date + datetime.timedelta(days=d + 1))
    return dates_f


def expand_dates(dates, T_forecast: int):
    """Extend existing dates list with forecast interval of length 'T_forecast'"""
    x_dates = dates.copy()
    for d in range(T_forecast):
        x_dates.append(dates[-1] + datetime.timedelta(days=d + 1))
    return x_dates


def prep_cases(raw_cases: pd.DataFrame, date_to_index: Optional[dict] = None):
    """Process raw_cases data to nd.array including unobserved dates.

    Parameters
    ----------
    raw_cases:
        a dataframe containing case counts with columns 'cases' and 'date'.

    date_to_index:
        optional dictionary for mapping calender dates to nd.array indices.

    Returns
    -------
    C:
        nd.array containing number of cases on each date.
    """
    raw_cases["date"] = pd.to_datetime(raw_cases["date"])
    if date_to_index is None:
        _, date_to_index = prep_dates(raw_cases["date"])

    T = len(date_to_index)
    C = np.full(T, np.nan)

    # Loop over dataframe rows to fill matrix
    for _, row in raw_cases.iterrows():
        C[date_to_index[row.date]] = row.cases
    return C


def format_var_names(raw_names: List[str], pivot: Optional[str] = None):
    """Places pivot category to be last element if present."""
    pivot = "other" if pivot is None else pivot
    if pivot in raw_names:
        # Move pivot the end of list
        names = raw_names.copy()
        names.append(pivot)
        names.remove(pivot)  # Removes first instance of pivot
        return names
    else:
        warnings.warn(
            f"{pivot} not present in variant names. Using provided order."
        )
    # Otherwise, return the original names
    return raw_names


def counts_to_matrix(
    raw_seqs: pd.DataFrame,
    var_names: List[str],
    date_to_index: Optional[dict] = None,
):
    """Process 'raw_seq' data to nd.array including unobserved dates.

    Parameters
    ----------
    raw_seq:
        a dataframe containing sequence counts with
        columns 'sequences' and 'date'.

    var_names:
        list of variant to count observations.

    date_to_index:
        optional dictionary for mapping calender dates to nd.array indices.

    Returns
    -------
    C:
        nd.array containing number of sequences of each variant on each date.
    """
    raw_seqs["date"] = pd.to_datetime(raw_seqs["date"])
    if date_to_index is None:
        _, date_to_index = prep_dates(raw_seqs["date"])

    T = len(date_to_index)
    V = len(var_names)
    C = np.full((T, V), np.nan)

    # Loop over rows in dataframe to fill matrix
    for i, s in enumerate(var_names):
        for _, row in raw_seqs[raw_seqs.variant == s].iterrows():
            C[date_to_index[row.date], i] = row.sequences

    # Loop over matrix rows to correct for zero counts
    for d in range(T):
        if not np.isnan(C[d, :]).all():
            # If not all days samples are nan, replace nan with zeros
            np.nan_to_num(C[d, :], copy=False)
    return C


def prep_sequence_counts(
    raw_seqs: pd.DataFrame,
    date_to_index: Optional[dict] = None,
    var_names: Optional[List] = None,
    pivot: Optional[str] = None,
):
    """Process 'raw_seq' data to nd.array including unobserved dates.

    Parameters
    ----------
    raw_seq:
        a dataframe containing sequence counts with
        columns 'sequences' and 'date'.


    date_to_index:
        optional dictionary for mapping calender dates to nd.array indices.

    var_names:
        optional list of variant to count observations.

    pivot:
        optional name of variant to place last.
        This will usually used as a reference or pivot strain.

    Returns
    -------
    var_names:
        list of variants counted

    C:
        nd.array containing number of sequences of each variant on each date.
    """

    if var_names is None:
        raw_var_names = list(pd.unique(raw_seqs.variant))
        raw_var_names.sort()
        var_names = format_var_names(raw_var_names, pivot=pivot)
    C = counts_to_matrix(raw_seqs, var_names, date_to_index=date_to_index)
    return var_names, C



from .LAS import LaplaceRandomWalk
from .basis_functions import *
from .model_options import *
from .model_helpers import *
from .renewal_model import *
from .spline_incidence import *
from .renewal_single_variant import *
from .renewal_regression import *



from functools import partial
from evofr.models.renewal_model.model_options import MultinomialSeq
from .model_spec import ModelSpec
import numpyro
import numpyro.distributions as dist
import numpy as np
import jax.numpy as jnp
from jax import jit, lax


def compute_frequency_piantham(ga, q0, gen_rev, T):
    """
    Compute variant frequencies according to Piantham model.

    Parameters
    ----------
    ga:
        Growth advantages for non-baseline variants.

    q0:
        Initial variant frequencies.

    gen_rev:
        Reversed generation time.

    T:
        Total length of time to simulate frequencies for.

    Returns
    -------
    Simulated frequencies as Array.
    """
    _ga = jnp.append(ga, 1.0)
    max_age = gen_rev.shape[-1]
    N_variants = q0.shape[-1]

    q0_padded = jnp.vstack((jnp.zeros((max_age - 1, N_variants)), q0))

    @jit
    def _scan_frequency(q, _):
        # Compute weighted frequency
        q_mag = _ga * jnp.einsum("dv, d -> v", q, gen_rev)
        q_new = q_mag / q_mag.sum()  # Normalize
        return jnp.vstack((q[-(max_age - 1) :, :], q_new)), q_new[:, None]

    _, q = lax.scan(_scan_frequency, init=q0_padded, length=T - 1, xs=None)
    return jnp.vstack((q0, jnp.squeeze(q)))


def Piantham_model_numpyro(
    seq_counts, N, gen_rev, SeqLik, forecast_L, pred=False
):
    T, N_variants = seq_counts.shape

    # Intial frequency
    q0 = numpyro.sample(
        "q0", dist.Dirichlet(jnp.ones(N_variants) / N_variants)
    )

    # Growth advantages
    with numpyro.plate("growth_advantage", N_variants - 1):
        ga = numpyro.sample("ga", dist.LogNormal(loc=0.0, scale=1.0))

    _T = T + forecast_L if pred else T
    _freq = compute_frequency_piantham(ga, q0, gen_rev, _T)

    freq = numpyro.deterministic(
        "freq", jnp.take(_freq, jnp.arange(T), axis=0)
    )
    numpyro.deterministic("s", ga - 1)

    # Compute likelihood of frequency
    SeqLik.model(seq_counts, N, freq, pred)

    if pred:
        numpyro.deterministic("freq_forecast", _freq[T:, :])


class PianthamModel(ModelSpec):
    """
    Model of frequency dynamics from Piantham 2021
    'Estimating the elevated transmissibility of the B.1.1.7
    strain over previously circulating strains in England
    using GISAID sequence frequencies'.
    """

    def __init__(self, gen, SeqLik=None, forecast_L=None):
        """Construct ModelSpec for frequency model with non-trivial generation time.

        Parameters
        ----------
        gen:
            Assumed generation time.

        SeqLik:
            Optional sequence likelihood option: MultinomialSeq or
            DirMultinomialSeq. Defaults to MultinomialSeq.

        forecast_L:
            Optional forecast length.

        Returns
        -------
        PianthamModel
        """
        self.gen = gen
        self.SeqLik = MultinomialSeq() if SeqLik is None else SeqLik
        self.forecast_L = 0 if forecast_L is None else forecast_L
        self.model_fn = partial(
            Piantham_model_numpyro,
            SeqLik=self.SeqLik,
            forecast_L=self.forecast_L,
        )

    def augment_data(self, data: dict) -> None:
        data["gen_rev"] = jnp.flip(self.gen, axis=-1)

        # Remove unnecessary key from VariantFrequencies
        data.pop("var_names", None)
        return None



from jax import random
import jax.numpy as jnp

from numpyro.infer import NUTS, MCMC, Predictive
from numpyro.infer.mcmc import MCMCKernel
from typing import Callable, Dict, Optional, Type


class MCMCHandler:
    def __init__(
        self,
        rng_key=1,
        kernel: Optional[Type[MCMCKernel]] = None,
        **kernel_kwargs
    ):
        """
        Construct MCMC handler.

        Parameters
        ----------
        rng_key:
            seed for pseudorandom number generator.

        kernel:
            optional MCMC kernel. Defaults to NUTS implementation in numpyro.

        **kernel_kwargs:
            kwargs to be passed to MCMC kernel at construction.

        Returns
        ------
        MCMCHandler
        """
        self.rng_key = random.PRNGKey(rng_key)
        self.kernel = kernel if kernel else NUTS
        self.kernel_kwargs = kernel_kwargs
        self.mcmc = None

    def fit(
        self,
        model: Callable,
        data: Dict,
        num_warmup: int,
        num_samples: int,
        **mcmc_kwargs
    ):
        """
        Fit model using MCMC given data.

        model:
            a numpyro model.

        data:
            dictionary containing arguments to 'model'.

        num_warmup:
            number of samples for warmup period in MCMC.

        num_samples:
            number of samples to be returned in MCMC.

        **mcmc_kwargs:
            additional arguments to be passed to MCMC algorithms.
        """
        self.mcmc = MCMC(
            self.kernel(model, **self.kernel_kwargs),
            num_warmup=num_warmup,
            num_samples=num_samples,
            **mcmc_kwargs
        )
        self.mcmc.run(self.rng_key, **data)
        self.samples = self.mcmc.get_samples()

    @property
    def params(self) -> Dict:
        if self.samples is not None:
            return self.samples
        return dict()

    def save_state(self, file_path):
        if self.samples is None:
            return None
        with open(file_path, "wb") as f:
            jnp.save(f, self.samples)

    def load_state(self, file_path):
        with open(file_path, "rb") as f:
            self.samples = jnp.load(f)

    def predict(self, model: Callable, data: Dict, **kwargs):
        predictive = Predictive(model, self.params)
        self.rng_key, rng_key_ = random.split(self.rng_key)
        samples_pred = predictive(rng_key_, pred=True, **data)
        return {**self.params, **samples_pred}



from typing import Optional, Type
from evofr.data.data_spec import DataSpec
from evofr.models.model_spec import ModelSpec
from evofr.posterior.posterior_handler import PosteriorHandler
import jax.numpy as jnp
from .SVI_handler import SVIHandler
from numpyro.infer import init_to_value
from numpyro.optim import Adam
from numpyro.infer.autoguide import AutoGuide
from numpyro.infer.autoguide import AutoDelta, AutoMultivariateNormal


class InferSVI:
    def __init__(
        self,
        iters: int,
        lr: float,
        num_samples: int,
        guide_fn: Type[AutoGuide],
    ):
        """Construct class for specifying SVI inference method.

        Parameters
        ----------
        iters:
            number of iterations to run optimizer.

        lr:
            learning rate for optimizer

        num_samples:
            number of samples to return from approximate posterior.

        guide_fn:
            variational model or guide to follow for SVI.

        Returns
        -------
        InferSVI
        """
        self.iters = iters
        self.num_samples = num_samples
        self.handler = SVIHandler(optimizer=Adam(lr))
        self.guide_fn = guide_fn

    def fit(
        self,
        model: ModelSpec,
        data: DataSpec,
        name: Optional[str] = None,
    ) -> PosteriorHandler:
        """Fit model given data using specificed SVI method.

        Parameters
        ----------
        model:
            ModelSpec for model

        data:
            DataSpec for data to do inference on

        name:
            name used to index posterior

        Returns
        -------
        PosteriorHandler
        """
        # Create and augment data dictionary
        input = data.make_data_dict()
        model.augment_data(input)

        # Create guide for SVI
        guide = self.guide_fn(model.model_fn)

        # Fit model and retrieve samples
        self.handler.fit(model.model_fn, guide, input, self.iters)
        samples = self.handler.predict(
            model.model_fn, guide, input, num_samples=self.num_samples
        )
        samples["losses"] = self.handler.losses

        # Create object to hold posterior samples and data
        self.posterior = PosteriorHandler(
            samples=samples, data=data, name=name if name is not None else ""
        )
        return self.posterior


class InferMAP(InferSVI):
    def __init__(self, iters: int, lr: float):
        super().__init__(iters, lr, 1, AutoDelta)


class InferFullRank(InferSVI):
    def __init__(self, iters: int, lr: float, num_samples: int):
        super().__init__(iters, lr, num_samples, AutoMultivariateNormal)


def init_to_MAP(
    model: ModelSpec,
    data: DataSpec,
    iters: int = 10_000,
    lr: float = 4e-3,
):
    """
    Initilization strategy for MCMC.
    Estimates MAP for given model and data.
    Returns initilization strategy and MAP estimates.
    """

    # Fit MAP with given model and data
    infer_map = InferMAP(iters=iters, lr=lr)
    MAP = infer_map.fit(model, data)

    # Return initilization strategy for MCMC and MAP estimates
    samples = {
        k: jnp.squeeze(v, axis=0) if v.shape[0] == 1 else v
        for k, v in MAP.samples.items()
    }
    return init_to_value(values=samples), MAP



import json
from typing import Dict, List, Optional

from evofr.data.data_spec import DataSpec
from evofr.posterior.posterior_helpers import EvofrEncoder


class PosteriorHandler:
    def __init__(
        self,
        samples: Optional[Dict] = None,
        data: Optional[DataSpec] = None,
        name: Optional[str] = None,
    ):
        """Construct PosteriorHandler.

        Parameters
        ----------
        samples:
            optional dictionary with keys corresponding to variable names.
            Values will be Arrays containing posterior samples.

        data:
            optional DataSpec instance containing underlying data from analysis

        name:
            name of the posterior. Used to index in MultiPosterior.
        """
        self.samples = samples if samples else dict()
        self.data = data
        self.name = name

    def save_posterior(self, filepath: str):
        """Save posterior samples at a given filepath."""
        if self.samples is not None:
            with open(filepath, "w") as file:
                json.dump(self.samples, file, cls=EvofrEncoder)

    def load_posterior(self, filepath: str):
        """Load posterior samples from a given filepath."""
        with open(filepath, "w") as file:
            self.samples = json.load(file)

    def unpack_posterior(self):
        """Return samples and dataspec from PosteriorHandler."""
        return self.samples, self.data

    def get_site(self, site: str):
        """Get samples at a given site from PosteriorHandler."""
        if self.samples:
            return self.samples[site]

    def get_data_dict(self):
        """Convert dataspec to corresponding dictionary."""
        if not self.data_dict and self.data:
            self.data_dict = self.data.make_data_dict()
        return self.data_dict


class MultiPosterior:
    def __init__(
        self,
        posteriors: Optional[List[PosteriorHandler]] = None,
        posterior: Optional[PosteriorHandler] = None,
    ):
        """Construct MultiPosterior.

        Parameters
        ----------
        posteriors:
            optional list of PosteriorHandlers to be added

        posterior:
            optional PosteriorHandler to be added to object.
        """
        self.locator = dict()
        if posterior is not None:
            self.add_posterior(posterior)
        if posteriors is not None:
            self.add_posteriors(posteriors)

    def add_posterior(self, posterior: PosteriorHandler):
        """Add PosteriorHandler to MultiPosterior by its name."""
        self.locator[posterior.name] = posterior

    def add_posteriors(self, posteriors: List[PosteriorHandler]):
        """Add multiple PosteriorHandlers to MultiPosterior by name."""
        for p in posteriors:
            self.add_posterior(p)

    def get(self, name):
        """Return PosteriorHandler by name"""
        return self.locator[name]

    def __getitem__(self, name: str):
        return self.get(name)



from typing import Callable, Optional
from jax import random, lax
from jax import jit
import jax.example_libraries.optimizers as optimizers

from numpyro.infer import SVI, Predictive, Trace_ELBO
from numpyro.infer.autoguide import AutoGuide
from numpyro.infer.svi import SVIState

import pickle


class SVIHandler:
    def __init__(
        self, rng_key=1, loss=Trace_ELBO(num_particles=2), optimizer=None
    ):
        """
        Construct SVI handler.

        Parameters
        ----------
        rng_key:
            seed for pseudorandom number generator.

        loss:
            optional loss to be used for MCMC.

        optimizer:
            optimizer to be used for SVI.

        Returns
        -------
        SVIHandler
        """

        self.rng_key = random.PRNGKey(rng_key)
        self.loss_fn = loss
        self.optimizer = optimizer
        self.svi_state = None

    def init_svi(self, model: Callable, guide: AutoGuide, data: dict):
        self.svi = SVI(model, guide, self.optimizer, self.loss_fn)
        svi_state = self.svi.init(self.rng_key, **data)
        if self.svi_state is None:
            self.svi_state = svi_state
        return self

    def fit(
        self, model: Callable, guide: AutoGuide, data: dict, n_epochs: int
    ):
        self.init_svi(model, guide, data)
        self.svi_result = self.svi.run(
            self.rng_key,
            n_epochs,
            **data,
            progress_bar=False,
            stable_update=True
        )
        self.svi_state = self.svi_result.state

    @property
    def params(self):
        if self.svi and self.svi_state:
            return self.svi.get_params(self.svi_state)

    @property
    def losses(self):
        if self.svi_result:
            return self.svi_result.losses

    def predict(self, model, guide, data, **kwargs):
        if self.svi is None:
            self.init_svi(model, guide, data)

        original_pred = Predictive(guide, params=self.params, **kwargs)
        self.rng_key, rng_key_ = random.split(self.rng_key)
        samples = original_pred(rng_key_, **data)
        predictive = Predictive(model, samples)

        self.rng_key, rng_key_ = random.split(self.rng_key)
        samples_pred = predictive(rng_key_, pred=True, **data)
        return {**samples, **samples_pred}

    def reset_state(self):
        return SVIHandler(self.rng_key, self.loss_fn, self.optimizer)

    # Optim state contains number of iterations and current state
    @property
    def optim_state(self):
        if self.svi_state is not None:
            return self.svi_state.optim_state

    def save_state(self, fp):
        with open(fp, "wb") as f:
            pickle.dump(
                optimizers.unpack_optimizer_state(self.optim_state[1]), f
            )

    def load_state(self, fp):
        with open(fp, "rb") as f:
            optim_state = (0, optimizers.pack_optimizer_state(pickle.load(f)))
        self.svi_state = SVIState(optim_state, None, self.rng_key)



from typing import Optional

from jax import random
from evofr.data.data_spec import DataSpec
from evofr.infer.InferBlackJax import BlackJaxNumpyro
from evofr.infer.backends import Backend
from evofr.models.model_spec import ModelSpec
from evofr.posterior.posterior_handler import PosteriorHandler


class SamplePrior:
    def __init__(self, num_samples: int, seed: Optional[int] = None):
        self.num_samples = num_samples
        self.seed = seed if seed is not None else 0
        self.rng_key = random.PRNGKey(self.seed)

    def _sample(
        self,
        model: ModelSpec,
        data: DataSpec,
        backend: Optional[Backend] = None,
    ):
        input = data.make_data_dict()
        model.augment_data(input)

        if hasattr(model, "backend") and backend is None:
            backend = model.backend

        # If no backend provided or defined elsewhere
        # default to numpyro
        if backend is None:
            return BlackJaxNumpyro.sample_prior(
                self.rng_key, model, input, self.num_samples
            )

        # Otherwise use provided backend
        if backend == Backend.NUMPYRO:
            return BlackJaxNumpyro.sample_prior(
                self.rng_key, model, input, self.num_samples
            )

    def sample(
        self, model: ModelSpec, data: DataSpec, name: Optional[str] = None
    ) -> PosteriorHandler:
        """
        Sample from model prior.
        """

        samples = self._sample(model, data)

        self.posterior = PosteriorHandler(
            samples=samples, data=data, name=name if name is not None else ""
        )
        return self.posterior



from .data_helpers import prep_dates, format_var_names
from .data_spec import DataSpec
from .case_counts import CaseCounts
from .variant_frequencies import VariantFrequencies
import numpy as np
import pandas as pd
from typing import List, Optional


class CaseFrequencyData(DataSpec):
    def __init__(
        self,
        raw_cases: pd.DataFrame,
        raw_seq: pd.DataFrame,
        date_to_index: Optional[dict] = None,
        var_names: Optional[List] = None,
        pivot: Optional[str] = None,
    ):
        """Construct a data specification for handling case counts
        and variant frequencies.

        Parameters
        ----------
        raw_cases:
            a dataframe containing case counts with columns 'cases' and 'date'.

        raw_seq:
            a dataframe containing sequence counts with columns 'sequences',
            'variant', and date'.

        date_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        var_names:
            optional list containing names of variants to be present

        pivot:
            optional name of variant to place last.
            Defaults to "other" if present otherwise.
            This will usually used as a reference or pivot strain.

        Returns
        -------
        CaseFrequencyData
        """

        # Get dates
        raw_dates = pd.concat((raw_cases["date"], raw_seq["date"]))
        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_dates)
        self.date_to_index = date_to_index

        # Get cases with CaseData
        case_data = CaseCounts(raw_cases, self.date_to_index)
        self.cases = case_data.cases

        # Get sequence data with VariantFrequencies
        freq_data = VariantFrequencies(
            raw_seq, self.date_to_index, var_names, pivot
        )
        self.seq_counts = freq_data.seq_counts
        self.var_names = freq_data.var_names
        self.pivot = self.var_names[-1]

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()

        data["cases"] = self.cases
        data["seq_counts"] = self.seq_counts
        data["N"] = self.seq_counts.sum(axis=1)
        data["var_names"] = self.var_names
        return data


class HierarchicalCFData:
    def __init__(
        self,
        raw_cases: pd.DataFrame,
        raw_seq: pd.DataFrame,
        group: str,
        date_to_index: Optional[dict] = None,
    ):
        """Construct a data specification for handling case counts
        and variant frequencies in a hierarchical model.

        Parameters
        ----------
        raw_cases:
            a dataframe containing case counts with columns 'cases' and 'date'.

        raw_seq:
            a dataframe containing sequence counts with columns 'sequences',
            'variant', and date'.

        data_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        group:
            string defining which column to seperate data by.

        Returns
        -------
        HierarchicalCFData
        """
        # Get dates
        raw_dates = raw_cases["date"].append(raw_seq["data"])
        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_dates)
        self.date_to_index = date_to_index

        # Get variant names
        raw_var_names = list(pd.unique(raw_seq.variant))
        raw_var_names.sort()
        self.var_names = format_var_names(raw_var_names)

        # Loop each group
        gseq = raw_seq.groupby(group)
        gcase = raw_cases.groupby(group)
        self.names = [name for name, _ in gseq]
        self.groups = [
            CaseFrequencyData(
                gcase.groups[name],
                gseq.groups[name],
                self.date_to_index,
                self.var_names,
            )
            for name in self.names
        ]

    def make_data_dict(self, data: Optional[dict] = None):
        if data is None:
            data = dict()

        data["cases"] = np.stack([g.cases for g in self.groups], axis=-1)
        data["seq_counts"] = np.stack(
            [g.seq_counts for g in self.groups],
            axis=-1,
        )
        data["N"] = np.stack(
            [g.seq_counts.sum(axis=-1) for g in self.groups], axis=-1
        )
        data["var_names"] = self.var_names
        return data



from .case_counts import CaseCounts
from .data_helpers import prep_dates
from .data_spec import DataSpec
import numpy as np
import pandas as pd
from typing import Optional


class HierCases(DataSpec):
    def __init__(
        self,
        raw_cases: pd.DataFrame,
        group: str,
        date_to_index: Optional[dict] = None,
    ):
        """Construct a data specification for handling case counts
        in hierarchical models.

        Parameters
        ----------
        raw_cases:
            a dataframe containing case counts with columns 'cases' and 'date'.

        group:
            string defining which column to seperate data by.

        date_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        Returns
        -------
        HierCases
        """
        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_cases["date"])
        self.date_to_index = date_to_index

        # Loop each group
        grouped = raw_cases.groupby(group)
        self.names = [name for name, _ in grouped]
        self.groups = [
            CaseCounts(group, self.date_to_index) for _, group in grouped
        ]

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()
        data["cases"] = np.stack([g.cases for g in self.groups], axis=-1)
        return data



from typing import Callable, Dict, Optional

import jax
from jax import random
from numpyro.infer.util import Predictive, initialize_model

import blackjax

from evofr.data.data_spec import DataSpec
from evofr.infer.backends import Backend
from evofr.models.model_spec import ModelSpec
from evofr.posterior.posterior_handler import PosteriorHandler


class BlackJaxNumpyro:
    @staticmethod
    def init(key, model, data):
        key, subkey = random.split(key)
        init_parms, potential_fn_gen, postprocess_fn, _ = initialize_model(
            subkey, model.model_fn, model_kwargs=data, dynamic_args=True
        )

        @jax.jit
        def logdensity_fn(position):
            return -potential_fn_gen(**data)(position)

        model.postprocess_fn = jax.jit(
            jax.vmap(postprocess_fn(**data), in_axes=0)
        )

        # Define initial position
        if hasattr(model, "initial_position"):
            initial_position = {
                site: model.initial_position[site]
                for site in init_parms.z.keys()
            }
            # initial_position = model.initial_position
        else:
            initial_position = init_parms.z

        return initial_position, logdensity_fn

    @staticmethod
    def predict(key, model, data, samples):
        samples = model.postprocess_fn(samples)
        predictive = Predictive(model.model_fn, samples)
        key, subkey = random.split(key)
        samples_pred = predictive(subkey, pred=True, **data)
        return {**samples, **samples_pred}

    @staticmethod
    def sample_prior(key, model, data, num_samples):
        key, subkey = random.split(key)
        prior_predictive = Predictive(model.model_fn, num_samples=num_samples)
        prior_samples = prior_predictive(subkey, pred=True, **data)
        return prior_samples


class BlackJaxProvided:
    @staticmethod
    def init(key, model, data: Dict):
        # Find density function with model_spec
        if hasattr(model, "logdensity_fn_gen"):
            logdensity_fn = model.logdensity_fn_gen(data)
        elif hasattr(model, "logdensity_fn"):
            logdensity_fn = model.logdensity_fn
        else:
            logdensity_fn = lambda _: None

        if hasattr(model, "initial_position"):
            init_position = model.initial_position
        elif hasattr(model, "initial_position_fn"):
            init_position = model.initial_position_fn(key)
        else:
            init_position = None
        return init_position, logdensity_fn

    @staticmethod
    def predict(key, model: ModelSpec, data: Dict, samples: Dict) -> Dict:
        if not hasattr(model, "pred_fn"):
            return samples
        samples_pred = model.pred_fn(key, samples, data)
        return {**samples, **samples_pred}


class BlackJaxHandler:
    def __init__(self, kernel, seed: Optional[int] = None, **kernel_kwargs):
        self.kernel_fn = kernel
        self.kernel_kwargs = kernel_kwargs
        self.seed = seed if seed is not None else 0
        self.rng_key = random.PRNGKey(self.seed)
        self.state = None

    @staticmethod
    def inference_loop(rng_key, kernel, initial_state, num_samples):
        @jax.jit
        def one_step(state, rng_key):
            state, info = kernel(rng_key, state)
            return state, (state, info)

        keys = random.split(rng_key, num_samples)
        _, (states, infos) = jax.lax.scan(one_step, initial_state, keys)
        return states, infos

    @staticmethod
    def _initialize(
        key,
        model: ModelSpec,
        data: Dict,
        backend: Optional[Backend] = None,
    ):
        # Check if model has a backend
        if hasattr(model, "backend") and backend is None:
            backend = model.backend

        # If no backend provided or defined elsewhere
        # default to numpyro
        if backend is None:
            return BlackJaxNumpyro.init(key, model, data)

        # Otherwise use provided backend
        if backend == Backend.NUMPYRO:
            return BlackJaxNumpyro.init(key, model, data)

        if backend == Backend.PROVIDED:
            return BlackJaxProvided.init(key, model, data)

        return None, lambda _: None

    @staticmethod
    def _predict(
        key,
        model: ModelSpec,
        data: Dict,
        samples: Dict,
        backend: Optional[Backend] = None,
    ):
        # Check if model has a backend
        if hasattr(model, "backend") and backend is None:
            backend = model.backend

        # Otherwise use suggested
        if backend == Backend.NUMPYRO:
            return BlackJaxNumpyro.predict(key, model, data, samples)

        if backend == Backend.PROVIDED:
            return BlackJaxProvided.predict(key, model, data, samples)

        # or default to numpyro
        if backend is None:
            return BlackJaxNumpyro.predict(key, model, data, samples)
        return dict()

    def run_warmup(
        self, initial_position, logdensity_fn: Callable, num_warmup: int
    ):
        num_warmup = 1 if num_warmup < 1 else num_warmup
        adapt = blackjax.window_adaptation(
            self.kernel_fn,
            logdensity_fn,
            **self.kernel_kwargs,
            num_steps=num_warmup
        )
        self.rng_key, key = random.split(self.rng_key)
        last_state, kernel, _ = adapt.run(key, initial_position)
        return last_state, kernel

    def fit(
        self,
        model: ModelSpec,
        data: Dict,
        num_warmup: int,
        num_samples: int,
    ):
        self.rng_key, key = random.split(self.rng_key)
        initial_position, logdensity_fn = self._initialize(key, model, data)

        # Run adapt window
        if num_warmup > 0:
            starting_state, kernel = self.run_warmup(
                initial_position, logdensity_fn, num_warmup=num_warmup
            )
        else:
            step_size = 1e-3
            inverse_mass_size = sum(
                v.size for _, v in initial_position.items()
            )
            inverse_mass_matrix = jax.numpy.ones(inverse_mass_size)
            kernel = self.kernel_fn(
                logdensity_fn, step_size, inverse_mass_matrix
            )
            starting_state = kernel.init(initial_position)
            kernel = kernel.step

        # Run sampling
        self.rng_key, key = random.split(self.rng_key)
        self.states, self.infos = self.inference_loop(
            key, kernel, starting_state, num_samples
        )

    @property
    def samples(self) -> Dict:
        if self.states is not None:
            return self.states.position
        return dict()

    def predict(self, model: ModelSpec, data: Dict) -> Dict:
        return self._predict(self.rng_key, model, data, self.samples)


class InferBlackJax:
    def __init__(
        self,
        num_warmup: int,
        num_samples: int,
        kernel,
        seed: Optional[int] = None,
        **kernel_kwargs
    ):
        """Construct class for specifying MCMC inference method.

        Parameters
        ----------
        num_warmup:
            number of warmup samples to run.

        num_samples:
            number of samples to return from MCMC.

        kernel:
            transition kernel for MCMC.

        Returns
        -------
        InferBlackJax
        """
        self.num_warmup = num_warmup
        self.num_samples = num_samples
        self.handler = BlackJaxHandler(
            kernel=kernel, seed=seed, **kernel_kwargs
        )

    def fit(
        self, model: ModelSpec, data: DataSpec, name: Optional[str] = None
    ) -> PosteriorHandler:
        """Fit model given data using specificed MCMC method.

        Parameters
        ----------
        model:
            ModelSpec for model

        data:
            DataSpec for data to do inference on

        name:
            name used to index posterior

        Returns
        -------
        PosteriorHandler
        """
        # Create and augment data dictionary
        input = data.make_data_dict()
        model.augment_data(input)

        # Fit model and retrieve samples
        self.handler.fit(model, input, self.num_warmup, self.num_samples)
        samples = self.handler.predict(model, input)

        # Create object to hold posterior samples and data
        self.posterior = PosteriorHandler(
            samples=samples, data=data, name=name if name is not None else ""
        )
        return self.posterior



from .model_spec import ModelSpec
from .multinomial_logistic_regression import MultinomialLogisticRegression
from .mlr_hierarchical import HierMLR
from .renewal_model import *
from .piantham_model import *
from .mlr_nowcast import *
from .mlr_spline import *
from .mlr_innovation import *



from typing import List, Optional
from jax._src.nn.functions import softmax
import jax.numpy as jnp
import numpy as np
from jax import jit, lax

from evofr.models.renewal_model.basis_functions.basis_fns import BasisFunction
from .LAS import LaplaceRandomWalk
import numpyro
import numpyro.distributions as dist
from .model_functions import reporting_to_vec
from .model_options import NegBinomCases, DirMultinomialSeq

from evofr.models.model_spec import ModelSpec
from .basis_functions import Spline


def renewal_regression_model_factory(
    g_rev,
    CaseLik=None,
    SeqLik=None,
):
    if CaseLik is None:
        CaseLik = NegBinomCases()
    if SeqLik is None:
        SeqLik = DirMultinomialSeq()

    def _variant_model(cases, seq_counts, N, X, var_names=None, pred=False):
        _, N_variant = seq_counts.shape
        T, k = X.shape

        # Make each variant have a growth advantage defined by a spline.
        # Might want to add control for the strength of gam_delta
        with numpyro.plate("N_variant_m1", N_variant - 1):
            delta_0 = numpyro.sample("delta_0", dist.Normal(0.0, 0.5))
            gam_delta = numpyro.sample("gam_delta", dist.HalfNormal(0.01))
            delta_rw = numpyro.sample(
                "delta_rw", LaplaceRandomWalk(scale=gam_delta, num_steps=k)
            )
            delta = delta_0 + delta_rw.T
        delta_mat = jnp.hstack((delta, jnp.zeros((k, 1))))
        freq = numpyro.deterministic(
            "freq", softmax(jnp.cumsum(jnp.dot(X, delta_mat), axis=0), axis=-1)
        )

        # Make total incidence defined by spline
        gam = numpyro.sample("gam", dist.HalfCauchy(0.1))

        beta_rw = numpyro.sample(
            "beta_rw", dist.GaussianRandomWalk(scale=gam, num_steps=k - 1)
        )
        beta_0 = numpyro.sample("beta_0", dist.Normal(0.0, 10.0))
        beta = numpyro.deterministic(
            "beta", beta_0 + jnp.concatenate([jnp.array([0.0]), beta_rw])
        )

        total_incidence = numpyro.deterministic(
            "total_smooth_prev", jnp.exp(jnp.dot(X, beta))
        )
        incidence = freq * total_incidence[:, None]

        # Compute reporting rate
        with numpyro.plate("rho_parms", 7):
            rho = numpyro.sample("rho", dist.Beta(5.0, 5.0))
        rho_vec = reporting_to_vec(rho, T)
        numpyro.deterministic("I_smooth", jnp.mean(rho_vec) * incidence)

        # Evaluate case likelihood
        CaseLik.model(cases, rho_vec * total_incidence, pred=pred)

        # Evaluate frequency likelihood
        SeqLik.model(seq_counts, N, freq, pred=pred)

        # Compute R and ga based on incidence above using generation time
        Rt, ga = rt_from_incidence(incidence, g_rev, T)

        numpyro.deterministic("R", Rt)
        numpyro.deterministic("ga", ga)
        numpyro.deterministic(
            "r", jnp.diff(jnp.log(incidence), prepend=0.0, axis=0)
        )

    return _variant_model


def rt_from_incidence(incidence, gen_rev, T):
    N_variants = incidence.shape[-1]
    max_age = gen_rev.shape[0]

    # Get indices of interest
    times = np.arange(T)
    ages = np.arange(max_age)
    gen_interval = ages + times[:, None]

    # Compute Rt
    inc_padded = jnp.vstack((jnp.zeros((max_age, N_variants)), incidence))
    infectivity = jnp.einsum("d, tdv -> tv", gen_rev, inc_padded[gen_interval])

    Rt = jnp.divide(incidence, infectivity)
    ga = jnp.divide(Rt, Rt[:, -1][:, None])
    return Rt, ga


class RenewalRegressionModel(ModelSpec):
    def __init__(
        self,
        gen,
        k: Optional[int] = None,
        CLik=None,
        SLik=None,
        v_names: Optional[List[str]] = None,
        basis_fn: Optional[BasisFunction] = None,
    ):
        self.gen = gen
        self.v_names = v_names

        # Making basis expansion for Rt
        self.k = k if k else 10
        self.basis_fn = (
            basis_fn if basis_fn else Spline(s=None, order=4, k=self.k)
        )

        self.CLik = CLik
        self.SLik = SLik
        self.make_model()

    def make_model(self):
        self.model_fn = renewal_regression_model_factory(
            jnp.flip(self.gen, axis=-1),
            self.CLik,
            self.SLik,
        )

    def augment_data(self, data, order=4):
        T = len(data["cases"])
        s = jnp.linspace(0, T, self.k)
        data["X"] = Spline.matrix(jnp.arange(T), s, order=order)



from typing import Dict, List, Optional, Tuple
import matplotlib as mpl
import matplotlib.pyplot as plt
from evofr.data.data_helpers import expand_dates
from evofr.data.data_spec import DataSpec
from evofr.plotting.plot_functions import (
    add_dates_sep,
    plot_R_censored,
    plot_cases,
    plot_ga_time_censored,
    plot_growth_advantage,
    plot_observed_frequency,
    plot_posterior_I,
    plot_posterior_frequency,
    plot_ppc_frequency,
    plot_time_varying_single,
    plot_time_varying_variant,
)

from evofr.posterior.posterior_handler import PosteriorHandler


DEFAULT_FIG_SIZE = (10, 6)
DEFAULT_PS = [0.8]
DEFAULT_ALPHAS = [0.4]


def create_empty_gridspec(
    nrows: int, ncols: int, figsize: Optional[Tuple] = None
):
    fig = plt.figure(
        figsize=figsize if figsize is not None else DEFAULT_FIG_SIZE
    )
    gs = fig.add_gridspec(nrows=nrows, ncols=ncols)
    return fig, gs


def create_date_axis(ax, plot_dates, date_sep=None, forecast_L=0):
    if forecast_L > 0:
        plot_dates = expand_dates(plot_dates, forecast_L)
        ax.axvline(
            x=len(plot_dates) - 1, color="k", linestyle="--"
        )  # Adding forecast cut off
    sep = date_sep if date_sep is not None else 20
    add_dates_sep(ax, plot_dates, sep=sep)  # Adding dates
    return None


def get_n_colors(n, cmap=None):
    """Returns a list of color of length n"""
    cmap = "tab20" if cmap is None else cmap
    cmap_fn = plt.cm.get_cmap(cmap, n)
    return [cmap_fn(i) for i in range(n)]


def get_colors(
    var_names: List[str],
    color_map: Optional[Dict] = None,
    cmap_name: Optional[str] = None,
):
    # If no color_map, generate based on color mapping
    if color_map is None:
        n_var = len(var_names)
        colors = get_n_colors(n_var, cmap_name)
        color_map = {v: c for v, c in zip(var_names, colors)}
    else:
        # Otherwise generate colors from provided mapping
        colors = [color_map[v] for v in var_names]
    return color_map, colors


class EvofrPlot:
    def __init__(
        self,
        posterior: Optional[PosteriorHandler] = None,
        samples: Optional[Dict] = None,
        data: Optional[DataSpec] = None,
    ):
        if posterior is not None:
            self.samples = posterior.samples
            self.data = posterior.data
        else:
            self.samples = samples if samples is not None else dict()
            self.data = data if data is not None else None


class FrequencyPlot(EvofrPlot):
    def __init__(
        self,
        posterior: Optional[PosteriorHandler] = None,
        samples: Optional[Dict] = None,
        data: Optional[DataSpec] = None,
        color_map: Optional[dict] = None,
        cmap_name: Optional[str] = None,
    ):
        super().__init__(posterior=posterior, samples=samples, data=data)
        self.color_map, self.colors = get_colors(
            self.data.var_names, color_map, cmap_name
        )

    def plot(
        self,
        ax=None,
        figsize: Optional[Tuple] = None,
        forecast: Optional[bool] = False,
        date_sep: Optional[int] = None,
        forecast_L: int = 0,
        posterior: bool = True,
        observed: bool = True,
        predictive: bool = False,
    ):
        if ax is None:
            # Create a figure and axis
            fig, gs = create_empty_gridspec(1, 1, figsize=figsize)
            ax = fig.add_subplot(gs[0])

        # Plot predicted frequencies
        if posterior:
            plot_posterior_frequency(
                ax,
                self.samples,
                DEFAULT_PS,
                DEFAULT_ALPHAS,
                self.colors,
                forecast=forecast,
            )

        if predictive:
            plot_ppc_frequency(
                ax,
                self.samples,
                self.data,
                DEFAULT_PS,
                DEFAULT_ALPHAS,
                self.colors,
            )

        if observed:
            plot_observed_frequency(
                ax, self.data, self.colors
            )  # Plot observed frequencies

        if hasattr(self.data, "dates"):
            create_date_axis(ax, self.data.dates, date_sep, forecast_L)
        ax.set_ylabel("Variant frequency")  # Making ylabel

        self.ax = ax
        return self


class GrowthAdvantagePlot(EvofrPlot):
    def __init__(
        self,
        posterior: Optional[PosteriorHandler] = None,
        samples: Optional[Dict] = None,
        data: Optional[DataSpec] = None,
        color_map: Optional[dict] = None,
        cmap_name: Optional[str] = None,
        time_varying: Optional[bool] = False,
        thres: Optional[float] = None,
        plot_pivot_line: Optional[bool] = True,
    ):
        super().__init__(posterior=posterior, samples=samples, data=data)
        self.color_map, self.colors = get_colors(
            self.data.var_names, color_map, cmap_name
        )
        self.thres = thres if thres is not None else 0.01
        self.time_varying = time_varying
        self.plot_pivot_line = plot_pivot_line

    def plot(
        self,
        ax=None,
        figsize: Optional[Tuple] = None,
        date_sep: Optional[int] = None,
    ):
        if ax is None:
            # Create a figure and axis
            fig, gs = create_empty_gridspec(1, 1, figsize=figsize)
            ax = fig.add_subplot(gs[0])

        if self.time_varying:
            plot_ga_time_censored(
                ax,
                self.samples,
                DEFAULT_PS,
                DEFAULT_ALPHAS,
                self.colors,
                thres=self.thres,
                plot_pivot_line=self.plot_pivot_line,
            )
            if hasattr(self.data, "dates"):
                create_date_axis(ax, self.data.dates, date_sep)
        else:
            plot_growth_advantage(
                ax,
                self.samples,
                self.data,
                DEFAULT_PS,
                DEFAULT_ALPHAS,
                self.colors,
                plot_pivot_line=self.plot_pivot_line,
            )
        ax.set_ylabel("Growth advantage")
        self.ax = ax
        return self


class RtPlot(EvofrPlot):
    def __init__(
        self,
        posterior: Optional[PosteriorHandler] = None,
        samples: Optional[Dict] = None,
        data: Optional[DataSpec] = None,
        color_map: Optional[dict] = None,
        cmap_name: Optional[str] = None,
        thres: Optional[float] = None,
        plot_neutral_line: Optional[bool] = True,
    ):
        super().__init__(posterior=posterior, samples=samples, data=data)
        self.color_map, self.colors = get_colors(
            self.data.var_names, color_map, cmap_name
        )
        self.thres = thres if thres is not None else 0.01
        self.plot_neutral_line = plot_neutral_line

    def plot(
        self,
        ax=None,
        figsize: Optional[Tuple] = None,
        date_sep: Optional[int] = None,
    ):
        # TODO: Add option to remove dashed line at 1.0
        if ax is None:
            # Create a figure and axis
            fig, gs = create_empty_gridspec(1, 1, figsize=figsize)
            ax = fig.add_subplot(gs[0])

        plot_R_censored(
            ax,
            self.samples,
            DEFAULT_PS,
            DEFAULT_ALPHAS,
            self.colors,
            thres=self.thres,
            plot_neutral_line=self.plot_neutral_line,
        )
        ax.set_ylabel("Effective Reproduction number")  # Making ylabel
        if hasattr(self.data, "dates"):
            create_date_axis(ax, self.data.dates, date_sep)

        self.ax = ax
        return self


class IncidencePlot(EvofrPlot):
    def __init__(
        self,
        posterior: Optional[PosteriorHandler] = None,
        samples: Optional[Dict] = None,
        data: Optional[DataSpec] = None,
        color_map: Optional[dict] = None,
        cmap_name: Optional[str] = None,
        color: str = "grey",
    ):
        super().__init__(posterior=posterior, samples=samples, data=data)
        self.color_map, self.colors = get_colors(
            self.data.var_names, color_map, cmap_name
        )
        self.color = color

    def plot(
        self,
        ax=None,
        figsize: Optional[Tuple] = None,
        cases: Optional[bool] = True,
        date_sep: Optional[int] = None,
        total: bool = False,
    ):
        if ax is None:
            # Create a figure and axis
            fig, gs = create_empty_gridspec(1, 1, figsize=figsize)
            ax = fig.add_subplot(gs[0])

        # Plot posterior variant specific incidence
        if total:
            plot_time_varying_single(
                ax,
                "total_smooth_prev",
                self.samples,
                DEFAULT_PS,
                DEFAULT_ALPHAS,
                self.color,
            )
        else:
            plot_posterior_I(
                ax,
                self.samples,
                DEFAULT_PS,
                DEFAULT_ALPHAS,
                self.colors,
            )

        if cases:
            plot_cases(ax, self.data)
        if hasattr(self.data, "dates"):
            create_date_axis(ax, self.data.dates, date_sep)
        self.ax = ax
        return self


class TimeVaryingPlot(EvofrPlot):
    def __init__(
        self,
        site: str,
        posterior: Optional[PosteriorHandler] = None,
        samples: Optional[Dict] = None,
        data: Optional[DataSpec] = None,
        by_variant: Optional[bool] = None,
        color_map: Optional[dict] = None,
        cmap_name: Optional[str] = None,
        color: Optional[str] = None,
    ):
        super().__init__(posterior=posterior, samples=samples, data=data)

        self.site = site

        if by_variant is None:
            by_variant = self.samples[self.site].ndim > 2

        if by_variant:
            self.color_map, self.colors = get_colors(
                self.data.var_names, color_map, cmap_name
            )
            self.plot_fn = plot_time_varying_variant
        else:
            self.colors = "grey" if color is None else color
            self.plot_fn = plot_time_varying_single

    def plot(
        self,
        ax=None,
        figsize: Optional[Tuple] = None,
        date_sep: Optional[int] = None,
    ):
        if ax is None:
            # Create a figure and axis
            fig, gs = create_empty_gridspec(1, 1, figsize=figsize)
            ax = fig.add_subplot(gs[0])
        self.plot_fn(
            ax,
            self.site,
            self.samples,
            DEFAULT_PS,
            DEFAULT_ALPHAS,
            self.colors,
        )
        if hasattr(self.data, "dates"):
            create_date_axis(ax, self.data.dates, date_sep)

        self.ax = ax
        return self


class PatchLegend:
    def __init__(
        self,
        color_map: Dict,
        ncol: Optional[int] = None,
        loc: Optional[str] = None,
    ):
        self.color_map = color_map
        self.ncol = ncol
        self.loc = loc

    def add_legend(self, fig=None, ax=None):
        if fig is None and ax is not None:
            fig = ax.get_figure()
        patches = [
            mpl.patches.Patch(color=color, label=label)
            for label, color in self.color_map.items()
        ]

        ncol = len(self.color_map.keys()) if self.ncol is None else self.ncol
        labels = list(self.color_map.keys())
        legend = fig.legend(patches, labels, ncol=ncol, loc=self.loc)
        legend.get_frame().set_linewidth(2.0)
        legend.get_frame().set_edgecolor("k")
        return fig



from typing import Optional, Type

from evofr.data.data_spec import DataSpec
from evofr.models.model_spec import ModelSpec
from .MCMC_handler import MCMCHandler
from evofr.posterior.posterior_handler import PosteriorHandler
from numpyro.infer import NUTS
from numpyro.infer.mcmc import MCMCKernel


class InferMCMC:
    def __init__(
        self,
        num_warmup: int,
        num_samples: int,
        kernel: Type[MCMCKernel],
        **kernel_kwargs
    ):
        """Construct class for specifying MCMC inference method.

        Parameters
        ----------
        num_warmup:
            number of warmup samples to run.

        num_samples:
            number of samples to return from MCMC.

        kernel:
            transition kernel for MCMC.

        Returns
        -------
        InferMCMC
        """
        self.num_warmup = num_warmup
        self.num_samples = num_samples
        self.handler = MCMCHandler(kernel=kernel, **kernel_kwargs)

    def fit(
        self, model: ModelSpec, data: DataSpec, name: Optional[str] = None
    ) -> PosteriorHandler:
        """Fit model given data using specificed MCMC method.

        Parameters
        ----------
        model:
            ModelSpec for model

        data:
            DataSpec for data to do inference on

        name:
            name used to index posterior

        Returns
        -------
        PosteriorHandler
        """
        # Create and augment data dictionary
        input = data.make_data_dict()
        model.augment_data(input)

        # Fit model and retrieve samples
        self.handler.fit(
            model.model_fn, input, self.num_warmup, self.num_samples
        )
        samples = self.handler.predict(model.model_fn, input)

        # Create object to hold posterior samples and data
        self.posterior = PosteriorHandler(
            samples=samples, data=data, name=name if name is not None else ""
        )
        return self.posterior


class InferNUTS(InferMCMC):
    def __init__(self, num_warmup: int, num_samples: int, **kernel_kwargs):
        super().__init__(num_warmup, num_samples, NUTS, **kernel_kwargs)



from typing import Optional
from jax import Array
import numpy as np
import jax.numpy as jnp
from jax.nn import softmax

import numpyro
import numpyro.distributions as dist

from evofr.models.renewal_model.basis_functions.basis_fns import BasisFunction
from evofr.models.renewal_model.basis_functions.splines import (
    Spline,
    SplineDeriv,
)

from .model_spec import ModelSpec


def MLR_spline_numpyro(
    seq_counts, N, X, X_deriv, tau=None, pred=False, var_names=None
):
    _, N_variants = seq_counts.shape
    _, N_features = X.shape

    # Sampling parameters
    raw_beta = numpyro.sample(
        "raw_beta",
        dist.Normal(0.0, 3.0),
        sample_shape=(N_features, N_variants - 1),
    )

    beta = numpyro.deterministic(
        "beta",
        jnp.column_stack(
            (raw_beta, jnp.zeros(N_features))
        ),  # All parameters are relative to last column / variant
    )

    logits = jnp.dot(X, beta)  # Logit frequencies by variant

    # Evaluate likelihood
    obs = None if pred else np.nan_to_num(seq_counts)
    numpyro.sample(
        "seq_counts",
        dist.MultinomialLogits(logits=logits, total_count=np.nan_to_num(N)),
        obs=obs,
    )

    # Compute frequency
    numpyro.deterministic("freq", softmax(logits, axis=-1))

    # Compute growth advantage from model
    if tau is not None:
        delta = jnp.dot(X_deriv, beta)
        numpyro.deterministic(
            "ga", jnp.exp(delta[:, :-1] * tau)
        )  # Last row corresponds to linear predictor / growth advantage


class MLRSpline(ModelSpec):
    def __init__(
        self,
        tau: float,
        s: Optional[Array] = None,
        k: Optional[int] = None,
        order: Optional[int] = None,
    ) -> None:
        """Construct ModelSpec for MLR with a spline basis.

        Parameters
        ----------
        tau:
            Assumed generation time for conversion to relative R.
        s:
            Array of knot locations for spline.
            Mutually exclusive argument with k.
        k:
            Number of basis elements.
            Mutually exclusive argument with s.
        order:
            Order of the spline to be used. Defaults to 4.
        Returns
        -------
        MLRSpline
        """
        self.tau = tau  # Fixed generation time
        self.s = s
        self.k = 10 if k is None or s is not None else k
        self.order = 4 if order is None else order
        self.basis_fn = Spline(s=self.s, order=self.order, k=self.k)
        self.basis_fn_deriv = SplineDeriv(s=self.s, order=self.order, k=self.k)
        self.model_fn = MLR_spline_numpyro

    def augment_data(self, data: dict) -> None:
        T = len(data["N"])
        data["X"] = self.basis_fn.make_features(data)
        data["X_deriv"] = self.basis_fn_deriv.make_features(data)
        data["tau"] = self.tau



from abc import ABC, abstractmethod
from typing import List, Optional
from jax._src.nn.functions import softmax
from jax import vmap
import numpy as np
import pandas as pd

from evofr.data.data_spec import DataSpec
from evofr.data.data_helpers import prep_dates, prep_sequence_counts
from evofr.models.renewal_model.model_options import MultinomialSeq
from .multinomial_logistic_regression import MultinomialLogisticRegression
import jax.numpy as jnp

import numpyro
import numpyro.distributions as dist

from .model_spec import ModelSpec

from functools import partial


class DeltaPriorModel(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def model(self):
        pass


class DeltaNormalPrior(DeltaPriorModel):
    def __init__(
        self, loc: Optional[float] = None, scale: Optional[float] = None
    ):
        self.loc = loc
        self.scale = scale

    def model(self, N_variants):
        delta_scale = (
            numpyro.sample("delta_scale", dist.HalfNormal(0.1))
            if self.loc is None
            else self.loc
        )
        delta_loc = (
            numpyro.sample("delta_loc", dist.Normal(0.0, 0.1))
            if self.scale is None
            else self.scale
        )
        raw_delta = numpyro.sample(
            "raw_delta",
            dist.Normal(delta_loc, delta_scale),
            sample_shape=(N_variants - 1,),
        )
        return raw_delta


class DeltaRegressionPrior(DeltaPriorModel):
    def __init__(self, features):
        """Construct a regression-based prior for relative fitness innovations.

        Parameters
        ----------
        features:
            Matrix of shape (N_variants, N_features)
            to predict relative fitness.

        Returns
        -------
        DeltaRegressionPrior
        """
        # Creature and process features (N_variants-1, n_features)
        self.features = features[:-1, :]
        self.N_features = self.features.shape[-1]
        self.is_missing = np.isnan(self.features).any(axis=1)
        self.n_missing = self.is_missing.sum()

    def model(self, N_variants):
        # Generate coefficients for features
        theta = numpyro.sample(
            "theta", dist.Normal(), sample_shape=(self.N_features,)
        )

        # Generate relative fitness where features present
        delta_loc_present = jnp.dot(self.features[~self.is_missing, :], theta)
        # Generate relative fitness where features missing
        delta_loc_missing = numpyro.sample(
            "delta_loc_missing",
            dist.Normal(0.0, 3.0),
            sample_shape=(self.n_missing,),
        )

        # Combine delta_loc for missing and present predictors
        delta_loc = jnp.empty(N_variants - 1)
        delta_loc = delta_loc.at[~self.is_missing].set(delta_loc_present)
        delta_loc = delta_loc.at[self.is_missing].set(delta_loc_missing)
        numpyro.deterministic("delta_loc", delta_loc)

        delta_scale = numpyro.sample("delta_scale", dist.HalfNormal(0.1))

        raw_delta = numpyro.sample(
            "raw_delta", dist.Normal(delta_loc, delta_scale)
        )
        return raw_delta

    def predict(self, features, samples):
        theta = samples["theta"]
        delta_scale = samples["delta_scale"]

        # Compute mean of delta
        delta_loc = jnp.einsum("vf, sf -> sv", features, theta)
        prior_delta = np.random.normal(delta_loc, delta_scale)
        return delta_loc, prior_delta


def MLR_innovation_model_time_varying(
    seq_counts, N, innovation_matrix, delta_prior, tau=None, pred=False
):
    T, N_variants = seq_counts.shape
    _, N_features = X.shape

    # Sampling parameters for growth advantages
    # Sampling intercepts
    raw_alpha = numpyro.sample(
        "raw_alpha",
        dist.Normal(0.0, 3.0),
        sample_shape=(N_variants - 1,),
    )
    alpha = numpyro.deterministic("alpha", jnp.append(raw_alpha, jnp.zeros(1)))

    # Now we need to get the time varying growth advantages
    # Sampling time-varying innovations from prior model
    raw_delta = delta_prior.model(N_variants) # (T, V)
    pivot_delta = jnp.dot(raw_delta,  innovation_matrix[-1,:-1])
    delta = numpyro.deterministic("delta", jnp.append(raw_delta, -pivot_delta))

    # Innovations to beta for growth advantages
    # (V, V) * (T, V) -> (T, V) # Gotta make sure shape right?
    mapped_dot = vmap(jnp.dot, in_axes=(None, 0), out_axes=0)
    beta = numpyro.deterministic("beta",  mapped_dot(innovation_matrix, delta))

    logits = alpha[:, None] + jnp.cumsum(beta, axis=0)
    numpyro.deterministic("freq", softmax(logits, axis=-1))

    # Evaluate likelihood of sequence counts given delays
    # SeqLik.model(seq_counts, N, freq, pred)  # Evaluate likelihood

    obs = None if pred else np.nan_to_num(seq_counts)
    numpyro.sample(
        "seq_counts",
        dist.Multinomial(total_count=np.nan_to_num(N), logits=logits),
        obs=obs,
    )

    # Compute growth advantage from model
    if tau is not None:
        numpyro.deterministic("ga", jnp.exp(beta * tau)[:-1])
        numpyro.deterministic("ga_delta", jnp.exp(delta * tau))


def MLR_innovation_model(
    seq_counts,
    N,
    X,
    innovation_matrix,
    delta_prior,
    tau=None,
    pred=False,
):
    T, N_variants = seq_counts.shape
    _, N_features = X.shape

    # Sampling parameters for growth advantages
    # Sampling intercepts
    raw_alpha = numpyro.sample(
        "raw_alpha",
        dist.Normal(0.0, 3.0),
        sample_shape=(N_variants - 1,),
    )
    raw_alpha = jnp.append(raw_alpha, jnp.zeros(1))

    # Sampling innovations from prior model
    raw_delta = delta_prior.model(N_variants)

    # We need last variant to have 0 growth advantage
    # so set its delta to negative of parent contributions
    pivot_delta = jnp.dot(raw_delta, innovation_matrix[-1, :-1])
    delta = jnp.append(raw_delta, -pivot_delta)
    numpyro.deterministic("delta", delta)

    # Innovations to beta for growth advantages
    raw_beta = jnp.dot(innovation_matrix, delta)

    beta = numpyro.deterministic(
        "beta", jnp.column_stack((raw_alpha, raw_beta)).T
    )

    logits = jnp.dot(X, beta)  # Logit frequencies by variant

    # Compute frequency
    # _freq = jnp.exp(logits)
    # _freq = jnp.exp(logits - jnp.max(logits, axis=-1, keepdims=True))
    numpyro.deterministic("freq", softmax(logits, axis=-1))

    # Evaluate likelihood of sequence counts given delays
    # SeqLik.model(seq_counts, N, freq, pred)  # Evaluate likelihood

    obs = None if pred else np.nan_to_num(seq_counts)
    numpyro.sample(
        "seq_counts",
        dist.Multinomial(total_count=np.nan_to_num(N), logits=logits),
        obs=obs,
    )

    # Compute growth advantage from model
    if tau is not None:
        numpyro.deterministic("ga", jnp.exp(raw_beta * tau)[:-1])
        numpyro.deterministic("ga_delta", jnp.exp(delta * tau))


class InnovationMLR(ModelSpec):
    def __init__(
        self, tau: float, delta_prior: Optional[DeltaPriorModel] = None
    ) -> None:
        """Construct ModelSpec for MLR allowing derived variants
        to share past fitness innovations.

        Parameters
        ----------
        tau:
            Assumed generation time for conversion to relative R.

        prior_model:
            Prior model for variant growtha advantage innovations.

        Returns
        -------
        InnovationMLR
        """
        self.tau = tau  # Fixed generation time
        self.delta_prior = (
            DeltaNormalPrior() if delta_prior is None else delta_prior
        )
        self.model_fn = partial(
            MLR_innovation_model, tau=self.tau, delta_prior=self.delta_prior
        )

    def augment_data(self, data: dict) -> None:
        T = len(data["N"])
        data["X"] = MultinomialLogisticRegression.make_ols_feature(
            0, T
        )  # Use intercept and time as predictors


def prep_clade_list(
    raw_variant_parents: pd.DataFrame,
    var_names: Optional[List] = None,
):
    """Process 'raw_variant_parents' data to nd.array showing
    which innovations present by variant.

    Parameters
    ----------
    raw_variant_parents:
        a dataframe containing variant name and parent variant

    var_names:
        optional list of variants

    pivot:
        optional name of variant to place last.
        Defaults to "other" if present otherwise.
        This will usually used as a reference or pivot strain.

    Returns
    -------
    innovation_matrix:
        binary matrix showing whether A[i,j] = i has innovation from j.
    """

    # First check:
    # SHould be able to reduce to a faster solve maybe?
    var_names = (
        list(raw_variant_parents.variant.unique())
        if var_names is None
        else var_names
    )
    innovation_matrix = np.zeros((len(var_names), len(var_names))).astype(int)
    parent_map = {
        row["variant"]: row["parent"]
        for _, row in raw_variant_parents.iterrows()
    }
    for _, row in raw_variant_parents.iterrows():
        var = row["variant"]
        if var not in var_names:
            continue
        var_index = var_names.index(var)

        # Find ancestors and add to row
        par = parent_map[var]
        is_par_variant = par in var_names
        while is_par_variant:
            par_index = var_names.index(par)  # Find index
            innovation_matrix[var_index, par_index] = 1  # Add innovation
            par = parent_map[par]  # New parent is parent of previous
            is_par_variant = par in var_names  # Check if should continue
        innovation_matrix[var_index, var_index] = 1
    return innovation_matrix.astype(bool), parent_map


class InnovationSequenceCounts(DataSpec):
    def __init__(
        self,
        raw_seq: pd.DataFrame,
        raw_variant_parents: pd.DataFrame,
        date_to_index: Optional[dict] = None,
        var_names: Optional[List] = None,
        pivot: Optional[str] = None,
    ):
        """Construct a data specification for handling variant frequencies.

        Parameters
        ----------
        raw_seq:
            a dataframe containing sequence counts with columns 'sequences',
            'variant', and date'.

        date_to_index:
            optional dictionary for mapping calender dates to nd.array indices.

        var_names:
            optional list containing names of variants to be present

        Returns
        -------
        InnovationSequenceCounts
        """

        if date_to_index is None:
            self.dates, date_to_index = prep_dates(raw_seq["date"])
        self.date_to_index = date_to_index

        # Turn dataframe to counts of each variant sequenced each day
        self.var_names, self.seq_counts = prep_sequence_counts(
            raw_seq, self.date_to_index, var_names, pivot=pivot
        )
        self.pivot = self.var_names[-1]

        # Create innovation_matrix
        self.innovation_matrix, self.parent_map = prep_clade_list(
            raw_variant_parents, self.var_names
        )

    def make_data_dict(self, data: Optional[dict] = None) -> dict:
        if data is None:
            data = dict()

        # Export `seq_counts`, sequences each day, and variant names
        data["seq_counts"] = self.seq_counts
        data["N"] = self.seq_counts.sum(axis=-1)
        data["innovation_matrix"] = self.innovation_matrix
        return data



from .backends import Backend
from .MCMC_handler import MCMCHandler
from .SVI_handler import SVIHandler
from .InferMCMC import InferMCMC, InferNUTS
from .InferSVI import InferSVI, InferMAP, InferFullRank, init_to_MAP
#from .InferBlackJax import InferBlackJax, BlackJaxHandler
#from .SamplePrior import *
